from __future__ import annotations

import urllib.parse
from typing import TYPE_CHECKING, Optional, Tuple, List, Any

if TYPE_CHECKING:
    from typing_extensions import Literal

import aiohttp

from .azure_exceptions import AzureRestApiError, raise_for_status
from .azure_identity import get_token
from .azure_rest_api import (
    azure_rest_api,
    azure_rest_api_paged,
    get_tenant_id,
)


async def get_acr_token(
    registry_name: str,
    token_type: Literal["refresh", "access"],
    method: Optional[str] = None,
    url_path: Optional[str] = None,
) -> str:
    """
    Gets tokens for interacting with the specified Azure Container Registry. Based on an
    incomplete understanding of OAuth:

    If token_type == "refresh", then this function will exchange the "default" Azure
    token from get_token for a container registry-specific refresh token and then return
    that.

    If token_type == "access", then this function will get the refresh token as
    described above, and then get a container registry-specific access token and return
    that. Getting an access token requires a scope, which we'll only have if
    method/url_path are provided. method/url_path should be what you will provide to
    azure_rest_api for the actual API call you want to make (i.e. url_path is everything
    after the domain name in the URL). We use this method/url_path to get a
    WWW-Authenticate header which will have a scope. (For refresh tokens, scopes aren't
    necessary, so we just query a generic endpoint that gives us the rest of the
    WWW-Authenticate header without the scope, which we won't need.)

    Confusingly, the original "default" Azure token appears to also be called an access
    token, but this is distinct from the container registry-specific access token
    returned by this function.
    """

    # References to the official Azure python SDK implementation
    # https://github.com/Azure/azure-cli/blob/284d6131679cecc26f928c78f1e4c29e2dd66405/src/azure-cli/azure/cli/command_modules/acr/_docker_utils.py#L115
    # https://github.com/Azure/azure-cli/blob/284d6131679cecc26f928c78f1e4c29e2dd66405/src/azure-cli-core/azure/cli/core/_profile.py#L349
    # https://github.com/Azure/azure-cli/blob/02cb00efbf661a1c18bab4a1fe03f2b1c59e30ef/src/azure-cli/azure/cli/command_modules/acr/_docker_utils.py#L365
    # https://github.com/Azure/azure-cli/blob/02cb00efbf661a1c18bab4a1fe03f2b1c59e30ef/src/azure-cli/azure/cli/command_modules/acr/_docker_utils.py#L403

    orig_token = await get_token()

    realm, service, scope = await _get_auth_parameters(registry_name, method, url_path)
    realm_url = urllib.parse.urlparse(realm)

    auth_host_exchange = urllib.parse.urlunparse(
        (realm_url[0], realm_url[1], "/oauth2/exchange", "", "", "")
    )
    headers = {"Content-Type": "application/x-www-form-urlencoded"}
    content_exchange = {
        # I think this refers to the type of token we're providing below, not the type
        # of token we're requesting
        "service": service,
        "tenant": await get_tenant_id(),
        "grant_type": "access_token",
        "access_token": orig_token,
    }
    async with aiohttp.request(
        "POST",
        auth_host_exchange,
        data=urllib.parse.urlencode(content_exchange),
        headers=headers,
    ) as response:
        response.raise_for_status()
        refresh_token = (await response.json())["refresh_token"]

    if token_type == "refresh":
        return refresh_token

    if scope is None:
        raise ValueError(
            "Requested an access token, but there was no scope in the WWW-Authenticate "
            "header. This probably means you did not pass in an appropriate url_path to"
            " _get_token"
        )
    auth_host_access = urllib.parse.urlunparse(
        (realm_url[0], realm_url[1], "/oauth2/token", "", "", "")
    )
    content_access = {
        "service": service,
        "scope": scope,
        "grant_type": "refresh_token",
        "refresh_token": refresh_token,
    }
    async with aiohttp.request(
        "POST",
        auth_host_access,
        data=urllib.parse.urlencode(content_access),
        headers=headers,
    ) as response:
        await raise_for_status(response)
        access_token = (await response.json())["access_token"]

    if token_type == "access":
        return access_token

    raise ValueError(f"Unexpected value for token_type: {token_type}")


async def _get_auth_parameters(
    registry_name: str, method: Optional[str] = None, url_path: Optional[str] = None
) -> Tuple[str, str, Optional[str]]:
    """
    Returns realm, service, scope. Basically just parses the WWW-Authenticate header
    based on method/url_path. Supplies a reasonable default method/url_path if None are
    provided.
    """

    # https://github.com/Azure/azure-cli/blob/284d6131679cecc26f928c78f1e4c29e2dd66405/src/azure-cli/azure/cli/command_modules/acr/_docker_utils.py#L63

    if method is None:
        method = "GET"
    if url_path is None:
        url_path = "v2/"

    async with aiohttp.request(
        method, f"https://{registry_name}.azurecr.io/{url_path}"
    ) as response:
        if response.status != 401:
            raise ValueError(
                f"Unexpected response from ACR login server, status {response.status}"
            )
        if "WWW-Authenticate" not in response.headers:
            raise ValueError(
                "Unexpected response from ACR login server, no WWW-Authenticate header"
            )
        authenticate = response.headers["WWW-Authenticate"]

    auth_type, sep, auth_parameters = authenticate.partition(" ")
    if sep != " ":
        raise ValueError(
            "Unexpected response from ACR login server, WWW-Authenticate header has no "
            "space"
        )
    if auth_type.lower() != "bearer":
        raise ValueError(
            "Unexpected response from ACR login server, WWW-Authenticate header does "
            f"not start with Bearer, instead starts with {auth_type}"
        )
    auth_parameters_dict = {}
    for pair in auth_parameters.split(","):
        key, sep, value = pair.partition("=")
        if sep != "=":
            raise ValueError(
                "Unexpected response from ACR login server. Auth parameters were not in"
                f" the expected format: {pair}"
            )
        auth_parameters_dict[key] = value.strip('"')

    if "realm" not in auth_parameters or "service" not in auth_parameters:
        raise ValueError(
            "Unexpected response from ACR login server, both realm and service auth "
            "parameter expected"
        )

    return (
        auth_parameters_dict["realm"],
        auth_parameters_dict["service"],
        auth_parameters_dict.get("scope"),
    )


# high-level ACR APIs


async def get_tags_in_repository(registry_name: str, repository: str) -> List[Any]:
    """
    Lists the tags in the specified repository. See
    https://docs.microsoft.com/en-us/rest/api/containerregistry/tag/get-list for
    documentation on what is returned. Returns a list of TagAttributesBase
    """
    method, url_path = "GET", f"acr/v1/{repository}/_tags"

    token = await get_acr_token(
        registry_name, "access", method=method, url_path=url_path
    )

    results = []
    try:
        async for page in azure_rest_api_paged(
            method,
            url_path,
            "",
            base_url=f"https://{registry_name}.azurecr.io",
            token=token,
        ):
            tags = page.get("tags")
            if tags:
                results.extend(tags)
    except AzureRestApiError as exn:
        # repository does not exist, i.e. no image has been pushed yet.
        if exn.status != 404:
            raise
    return results


async def delete_tag(registry_name: str, repository: str, tag: str) -> None:
    """
    Deletes the specified tag. See
    https://docs.microsoft.com/en-us/rest/api/containerregistry/tag/delete
    """
    method, url_path = "DELETE", f"acr/v1/{repository}/_tags/{tag}"

    token = await get_acr_token(
        registry_name, "access", method=method, url_path=url_path
    )

    await azure_rest_api(
        method,
        url_path,
        "2021-07-01",
        base_url=f"https://{registry_name}.azurecr.io",
        token=token,
    )



import asyncio
from typing import Tuple, Optional

import asyncssh
from cryptography.hazmat.primitives import serialization as crypto_serialization
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey

from meadowrun.azure_integration.azure_meadowrun_core import (
    assign_role_to_principal,
    ensure_meadowrun_resource_group,
    get_subscription_id,
)
from meadowrun.azure_integration.mgmt_functions.azure_core.azure_exceptions import (
    ResourceNotFoundError,
)
from meadowrun.azure_integration.mgmt_functions.azure_core.azure_identity import (
    get_current_user_id,
)
from meadowrun.azure_integration.mgmt_functions.azure_core.azure_rest_api import (
    azure_rest_api,
    azure_rest_api_paged,
    azure_rest_api_poll,
    get_tenant_id,
    wait_for_poll,
)
from meadowrun.ssh_keys import generate_rsa_private_key

_MEADOWRUN_PRIVATE_SSH_KEY_SECRET_NAME = "meadowrunSshPrivateKey"


def get_meadowrun_vault_name(subscription_id: str) -> str:
    # key vault names need to be globally unique. Also, they cannot be more than 24
    # characters long
    # TODO we need a way to manually set Key Vault name in case this ends up colliding
    return "mr" + subscription_id.replace("-", "")[-22:]


async def _ensure_meadowrun_vault(location: str) -> str:
    """
    Gets the meadowrun key vault URI if it exists. If it doesn't exist, also creates the
    meadowrun key vault, and tries to assign the Key Vault Administrator role to the
    current user.
    """

    subscription_id = await get_subscription_id()
    vault_name = get_meadowrun_vault_name(subscription_id)
    resource_group_path = await ensure_meadowrun_resource_group(location)
    vault_path = (
        f"{resource_group_path}/providers/Microsoft.KeyVault/vaults/{vault_name}"
    )

    try:
        vault = await azure_rest_api("GET", vault_path, "2019-09-01")
        return vault["properties"]["vaultUri"]
    except ResourceNotFoundError:
        # theoretically key_vault_client.vaults.get_deleted() should be faster,
        # but that requires specifying a location and there's no way to know what
        # location the key vault may have been originally created in.
        deleted_vault_found = False
        async for page in azure_rest_api_paged(
            "GET",
            f"/subscriptions/{subscription_id}/providers/Microsoft.KeyVault/"
            f"deletedVaults",
            "2019-09-01",
        ):
            for vault in page["value"]:
                if vault["name"] == vault_name:
                    deleted_vault_found = True
                    break

            if deleted_vault_found:
                break

        if deleted_vault_found:
            # if we have a deleted vault, then we should try to recover it
            create_mode = "recover"
            print(f"The meadowrun Key Vault {vault_name} was deleted, recovering")
        else:
            create_mode = "default"
            print(
                f"The meadowrun Key Vault {vault_name} does not exist, creating it "
                "now"
            )

        # if we're creating or recreating the Key Vault, assume that we need to add the
        # current user to the Key Vault Administrator role so that the current user can
        # access secrets.
        assign_role_task = asyncio.create_task(
            assign_role_to_principal(
                "Key Vault Administrator", await get_current_user_id(), location
            )
        )

        # Now we can create/recover the Key Vault.
        # https://docs.microsoft.com/en-us/rest/api/keyvault/keyvault/vaults/create-or-update#vaultproperties
        vault, _ = await wait_for_poll(
            await azure_rest_api_poll(
                "PUT",
                vault_path,
                "2019-09-01",
                "AsyncOperationJsonStatus",
                json_content={
                    "location": location,
                    "properties": {
                        "tenantId": await get_tenant_id(),
                        "sku": {"family": "A", "name": "Standard"},
                        "enableRbacAuthorization": True,
                        "createMode": create_mode,
                    },
                },
            )
        )

        try:
            await assign_role_task
        except Exception as e:
            print(
                "Warning: we were not able to assign the Key Vault Administrator role "
                f"to the current user. You may not be able to create/read secrets: {e}"
            )

        return vault["properties"]["vaultUri"]


_private_key: Optional[asyncssh.SSHKey] = None


async def ensure_meadowrun_key_pair(location: str) -> Tuple[asyncssh.SSHKey, str]:
    """
    Makes sure that a private key is stored in a Key Vault secret. If it does not exist,
    creates a fresh private key and uploads it.

    Returns (private key, public key). Private key is an asyncssh.SSHKey that can be
    passed to ssh.connect(private_key=pkey), public key is a string that can be passed
    into launch_vms.
    """
    global _private_key

    vault_uri = await _ensure_meadowrun_vault(location)

    if _private_key is None:
        try:
            result = await azure_rest_api(
                "GET",
                f"secrets/{_MEADOWRUN_PRIVATE_SSH_KEY_SECRET_NAME}",
                "7.3",
                base_url=vault_uri,
                token_scope="https://vault.azure.net/.default",
            )
            private_key_text = result["value"]
        except ResourceNotFoundError:
            try:
                # if we have a deleted secret, recover it, as we cannot add update a
                # deleted secret, we must recover it first. No need to check if the
                # deleted secret exists already, this call will just fail with
                # ResourceNotFoundError
                await azure_rest_api(
                    "POST",
                    f"deletedsecrets/{_MEADOWRUN_PRIVATE_SSH_KEY_SECRET_NAME}/recover",
                    "7.3",
                    base_url=vault_uri,
                    token_scope="https://vault.azure.net/.default",
                )
            except ResourceNotFoundError:
                pass

            print(
                "The Azure Key Vault secret that stores the meadowrun private key does "
                "not exist, creating it now"
            )

            private_key_text = generate_rsa_private_key()
            # create the private key secret
            # https://docs.microsoft.com/en-us/rest/api/keyvault/secrets/set-secret/set-secret
            await azure_rest_api(
                "PUT",
                f"secrets/{_MEADOWRUN_PRIVATE_SSH_KEY_SECRET_NAME}",
                "7.3",
                base_url=vault_uri,
                token_scope="https://vault.azure.net/.default",
                json_content={"value": private_key_text},
            )

        _private_key = asyncssh.import_private_key(private_key_text)

    rsa_key = _private_key.pyca_key
    if not isinstance(rsa_key, RSAPrivateKey):
        raise ValueError(
            "Private key text is bad, the resulting RSAKey.key is of type "
            f"{type(rsa_key)}"
        )
    return _private_key, rsa_key.public_key().public_bytes(
        crypto_serialization.Encoding.OpenSSH, crypto_serialization.PublicFormat.OpenSSH
    ).decode("utf-8")


async def download_ssh_key(output_path: str, location: str) -> None:
    """Downloads the meadowrun private SSH key to the specified file name"""
    vault_uri = await _ensure_meadowrun_vault(location)
    result = await azure_rest_api(
        "GET",
        f"secrets/{_MEADOWRUN_PRIVATE_SSH_KEY_SECRET_NAME}",
        "7.3",
        base_url=vault_uri,
        token_scope="https://vault.azure.net/.default",
    )
    key = result["value"]

    if key is None:
        raise ValueError("Azure SSH key secret's value was None")

    with open(output_path, "w", encoding="utf-8") as output_file:
        output_file.write(key)



from __future__ import annotations

import itertools
import json
import os
import time
import traceback
from typing import (
    TYPE_CHECKING,
    Any,
    Iterable,
    List,
    Optional,
    Tuple,
    TypeVar,
)

import aiobotocore.session

from meadowrun.aws_integration.management_lambdas.ec2_alloc_stub import (
    _EC2_ALLOC_TAG,
    _EC2_ALLOC_TAG_VALUE,
)
from meadowrun.meadowrun_pb2 import ProcessState
from meadowrun.storage_grid_job import (
    S3Bucket,
    complete_task,
    download_task_arg,
    get_aws_s3_bucket,
    upload_task_args,
)
from meadowrun.run_job_local import restart_worker

if TYPE_CHECKING:
    from types_aiobotocore_sqs.client import SQSClient
    from meadowrun.run_job_local import TaskWorkerServer, WorkerMonitor

_REQUEST_QUEUE_NAME_PREFIX = "meadowrun-task-"

_T = TypeVar("_T")


async def create_request_queue(job_id: str, sqs: SQSClient) -> str:
    """
    See create_queues_and_add_tasks

    Mote that queue names can't contain anything other than letters, numbers, and
    hyphens
    """

    # interestingly, this call will not fail if the exact same queue already exists,
    # but if the queue exists but has different attributes/tags, it will throw an
    # exception
    # TODO try to detect job id collisions?
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sqs.html#SQS.Client.create_queue
    request_queue_response = await sqs.create_queue(
        QueueName=f"{_REQUEST_QUEUE_NAME_PREFIX}{job_id}",
        tags={_EC2_ALLOC_TAG: _EC2_ALLOC_TAG_VALUE},
    )

    # FIFO SQS queues would be nice because they give an "at most once" message delivery
    # guarantee, but we don't need the ordering guarantee, and the performance penalty
    # introduced by the ordering guarantee is unacceptable. It's possible we'll end up
    # executing a task more than once, but according to the docs this should be very
    # rare. If it does happen, then we'll overwrite the result, which is fine.
    # TODO We should put in a check to notify the user that a task ran more than once

    request_queue_url = request_queue_response["QueueUrl"]

    return request_queue_url


def _chunker(it: Iterable[_T], size: int) -> Iterable[Tuple[_T, ...]]:
    """E.g. _chunker([1, 2, 3, 4, 5], 2) -> [1, 2], [3, 4], [5]"""
    iterator = iter(it)
    while True:
        chunk = tuple(itertools.islice(iterator, size))
        if not chunk:
            break
        yield chunk


MESSAGE_PREFIX_TASK = "task"
MESSAGE_PREFIX_WORKER_SHUTDOWN = "worker-shutdown"


async def add_tasks(
    job_id: str,
    request_queue_url: str,
    s3_bucket: S3Bucket,
    sqs: SQSClient,
    run_map_args: Iterable[Any],
) -> List[Tuple[int, int]]:
    """
    This can only be called once per request_queue_url. If we wanted to support calling
    add_tasks more than once for the same job, we would need the caller to manage the
    task_ids
    """

    byte_ranges = await upload_task_args(s3_bucket, job_id, run_map_args)

    for byte_ranges_chunk in _chunker(enumerate(byte_ranges), 10):
        # this function can only take 10 messages at a time, so we chunk into
        # batches of 10
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sqs.html#SQS.Client.send_message_batch

        entries = [
            {
                "Id": str(i),
                "MessageBody": json.dumps(
                    {
                        "task_id": i,
                        "attempt": 1,
                        "range_from": range_from,
                        "range_end": range_to,
                    }
                ),
            }
            for i, (range_from, range_to) in byte_ranges_chunk
        ]

        result = await sqs.send_message_batch(
            QueueUrl=request_queue_url, Entries=entries  # type: ignore
        )
        if "Failed" in result:
            raise ValueError(f"Some grid tasks could not be queued: {result['Failed']}")

    return byte_ranges


async def retry_task(
    request_queue_url: str,
    task_id: int,
    attempt: int,
    byte_range: Tuple[int, int],
    sqs_client: SQSClient,
) -> None:
    task = json.dumps(
        dict(
            task_id=task_id,
            attempt=attempt,
            range_from=byte_range[0],
            range_end=byte_range[1],
        )
    )

    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sqs.html#SQS.Client.send_message
    await sqs_client.send_message(QueueUrl=request_queue_url, MessageBody=task)


async def add_worker_shutdown_messages(
    request_queue_url: str,
    num_messages_to_send: int,
    sqs_client: SQSClient,
) -> None:
    messages_sent = 0
    while messages_sent < num_messages_to_send:
        num_messages_to_send_batch = min(10, num_messages_to_send - messages_sent)
        # the id only needs to be unique within a single request
        entries = [
            {"Id": str(i), "MessageBody": "{}"}
            for i in range(num_messages_to_send_batch)
        ]

        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sqs.html#SQS.Client.send_message
        result = await sqs_client.send_message_batch(
            QueueUrl=request_queue_url, Entries=entries  # type: ignore
        )
        if "Failed" in result:
            raise ValueError(
                "Some worker shutdown messages could not be queued: "
                f"{result['Failed']}"
            )

        messages_sent += num_messages_to_send_batch


_GET_TASK_TIMEOUT_SECONDS = 60 * 2  # 2 minutes


async def _get_task(
    sqs: SQSClient,
    s3_bucket: S3Bucket,
    request_queue_url: str,
    job_id: str,
    receive_message_wait_seconds: int,
) -> Optional[Tuple[int, int, bytes]]:
    """
    Gets the next task from the specified request_queue, downloads the task argument
    from S3, and then deletes the message from the request_queue.

    Returns a tuple of task_id and pickled argument if there was a task, otherwise
    returns None. Waits receive_message_wait_seconds for a task.
    """
    # get the task message
    t0 = time.time()
    while True:
        if time.time() - t0 > _GET_TASK_TIMEOUT_SECONDS:
            raise TimeoutError(
                f"Waited more than {_GET_TASK_TIMEOUT_SECONDS} for the next task but no"
                " task was available. This is unexpected--the GridJobDriver should have"
                " sent a shutdown message or a SIGINT explicitly"
            )

        result = await sqs.receive_message(
            QueueUrl=request_queue_url, WaitTimeSeconds=receive_message_wait_seconds
        )

        if "Messages" in result:
            break

    messages = result["Messages"]
    if len(messages) != 1:
        raise ValueError(f"Requested one message but got {len(messages)}")

    # parse the task request
    task = json.loads(messages[0]["Body"])
    if "task_id" in task:
        task_id, attempt, range_from, range_end = (
            task["task_id"],
            task["attempt"],
            task["range_from"],
            task["range_end"],
        )
        arg = await download_task_arg(s3_bucket, job_id, (range_from, range_end))

        # TODO store somewhere (DynamoDB?) that this worker has picked up the task.

        # acknowledge receipt/delete the task request message so we don't have duplicate
        # tasks running
        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sqs.html#SQS.Client.delete_message
        await sqs.delete_message(
            QueueUrl=request_queue_url, ReceiptHandle=messages[0]["ReceiptHandle"]
        )

        return task_id, attempt, arg

    # it's a worker exit message, so return None to exit
    await sqs.delete_message(
        QueueUrl=request_queue_url, ReceiptHandle=messages[0]["ReceiptHandle"]
    )
    return None


async def _worker_iteration(
    sqs: SQSClient,
    s3_bucket: S3Bucket,
    request_queue_url: str,
    job_id: str,
    log_file_name: str,
    pid: int,
    worker_server: TaskWorkerServer,
    worker_monitor: WorkerMonitor,
) -> bool:
    task = await _get_task(
        sqs,
        s3_bucket,
        request_queue_url,
        job_id,
        3,
    )
    if not task:
        print("Meadowrun agent: Received shutdown message. Exiting.")
        return False

    task_id, attempt, arg = task
    print(f"Meadowrun agent: About to execute task #{task_id}, attempt #{attempt}")
    try:
        worker_monitor.start_stats()
        await worker_server.send_message(arg)

        state, result = await worker_server.receive_message()
        stats = await worker_monitor.stop_stats()

        process_state = ProcessState(
            state=ProcessState.ProcessStateEnum.SUCCEEDED
            if state == "SUCCEEDED"
            else ProcessState.ProcessStateEnum.PYTHON_EXCEPTION,
            pid=pid,
            pickled_result=result,
            return_code=0,
            log_file_name=log_file_name,
            max_memory_used_gb=stats.max_memory_used_gb,
        )

    except Exception:
        stats = await worker_monitor.stop_stats()

        process_state = ProcessState(
            state=ProcessState.ProcessStateEnum.UNEXPECTED_WORKER_EXIT,
            pid=pid,
            return_code=(await worker_monitor.try_get_return_code()) or 0,
            log_file_name=log_file_name,
            max_memory_used_gb=stats.max_memory_used_gb,
        )

        print(
            f"Meadowrun agent: Unexpected worker exit, restarting worker. "
            f"In task #{task_id}, attempt #{attempt}, {traceback.format_exc()}"
        )

        await restart_worker(worker_server, worker_monitor)

    print(
        f"Meadowrun agent: Completed task #{task_id}, attempt #{attempt}, "
        f"state {ProcessState.ProcessStateEnum.Name(process_state.state)}, max "
        f"memory {process_state.max_memory_used_gb}GB "
    )

    await complete_task(s3_bucket, job_id, task_id, attempt, process_state)

    return True


async def worker_function(
    request_queue_url: str,
    job_id: str,
    region_name: str,
    public_address: str,
    log_file_name: str,
    worker_server: TaskWorkerServer,
    worker_monitor: WorkerMonitor,
) -> None:
    """
    Runs a loop that gets tasks off of the request_queue, communicates that via reader
    and writer to the task worker, and uploads the results to S3.
    """
    pid = os.getpid()
    log_file_name = f"{public_address}:{log_file_name}"
    session = aiobotocore.session.get_session()
    async with session.create_client(
        "sqs", region_name=region_name
    ) as sqs, get_aws_s3_bucket(region_name) as s3_bucket:
        await worker_server.wait_for_task_worker_connection()
        while await _worker_iteration(
            sqs,
            s3_bucket,
            request_queue_url,
            job_id,
            log_file_name,
            pid,
            worker_server,
            worker_monitor,
        ):
            pass



import asyncio.subprocess
import dataclasses
import datetime
import time
from typing import (
    Any,
    AsyncIterator,
    Coroutine,
    Dict,
    Iterable,
    Optional,
    Tuple,
    cast,
)
from typing_extensions import Literal

import aiohttp
import requests

from .azure_exceptions import AzureRestApiError, raise_for_status, raise_for_status_sync
from .azure_identity import CACHED_AZURE_VARIABLES, get_token, get_token_sync

_BASE_URL = "https://management.azure.com"


async def _return_response(response: aiohttp.ClientResponse) -> Any:
    # If there's nothing to return, the Azure APIs will return content-type:
    # application/octet-stream with no actual content. It's easier to just return this
    # as None rather than having to fully support another content type
    if response.content_length == 0:
        return None

    # this is a bit sloppy, but sometimes content-length is not set, content-type is
    # text/plain, and the actual content is empty. We don't want response.json() to
    # throw in that scenario either
    if response.content_type.startswith("text/plain"):
        return await response.text()

    if response.content_type.startswith("application/xml"):
        return await response.text()

    if response.content_type.startswith("application/json"):
        return await response.json()

    # assuming binary
    return await response.content.read()


def _prepare_request(
    method: str,
    url_path: str,
    api_version: str,
    token: str,
    *,
    query_parameters: Optional[Dict[str, str]] = None,
    json_content: Any = None,
    base_url: Optional[str] = None,
) -> Dict[str, Any]:
    if not base_url:
        base_url = _BASE_URL
    url = f"{base_url}/{url_path}"
    headers = {"Authorization": f"Bearer {token}"}
    parameters = {"api-version": api_version}
    if query_parameters:
        parameters.update(query_parameters)

    return {
        "method": method,
        "url": url,
        "params": parameters,
        "headers": headers,
        "json": json_content,
    }


async def azure_rest_api(
    method: str,
    url_path: str,
    api_version: str,
    *,
    query_parameters: Optional[Dict[str, str]] = None,
    json_content: Any = None,
    base_url: Optional[str] = None,
    token: Optional[str] = None,
    token_scope: Optional[str] = None,
) -> Any:
    """
    Supports any Azure REST API call, returns the JSON-deserialized content of the
    response
    """

    if token is None:
        token = await get_token(token_scope)
    async with aiohttp.request(
        **_prepare_request(
            method,
            url_path,
            api_version,
            token,
            query_parameters=query_parameters,
            json_content=json_content,
            base_url=base_url,
        )
    ) as response:
        await raise_for_status(response)
        return await _return_response(response)


async def azure_rest_api_paged(
    method: str,
    url_path: str,
    api_version: str,
    *,
    query_parameters: Optional[Dict[str, str]] = None,
    base_url: Optional[str] = None,
    token: Optional[str] = None,
) -> AsyncIterator[Any]:
    """
    Supports any Azure REST API call that takes a $maxpagesize parameter and returns a
    json body with a nextLink parameter. Returns the json-deserialized body of each
    call. Typical usage will be [item async for page in azure_rest_api_paged(...) for
    item in page["value"]]
    """
    if token is None:
        token = await get_token()
    prepared_request = _prepare_request(
        method,
        url_path,
        api_version,
        token,
        query_parameters=query_parameters,
        base_url=base_url,
    )
    prepared_request["params"].setdefault("$maxpagesize", "1000")

    async with aiohttp.request(**prepared_request) as response:
        await raise_for_status(response)
        response_json = await response.json()
        next_link = response_json.get("nextLink")
        yield response_json

    while next_link:
        async with aiohttp.request(
            method, next_link, headers=prepared_request["headers"]
        ) as response:
            await raise_for_status(response)
            response_json = await response.json()
            next_link = response_json.get("nextLink")
            yield response_json


def azure_rest_api_paged_sync(
    method: str,
    url_path: str,
    api_version: str,
    *,
    query_parameters: Optional[Dict[str, str]] = None,
    base_url: Optional[str] = None,
    token: Optional[str] = None,
) -> Iterable[Any]:
    if token is None:
        token = get_token_sync()
    prepared_request = _prepare_request(
        method,
        url_path,
        api_version,
        token,
        query_parameters=query_parameters,
        base_url=base_url,
    )
    prepared_request["params"].setdefault("$maxpagesize", "1000")

    response = requests.request(**prepared_request)
    raise_for_status_sync(response)
    response_json = response.json()
    next_link = response_json.get("nextLink")
    yield response_json

    while next_link:
        response = requests.request(
            method, next_link, headers=prepared_request["headers"]
        )
        raise_for_status_sync(response)
        response_json = response.json()
        next_link = response_json.get("nextLink")
        yield response_json


DEFAULT_TOTAL_POLL_TIMEOUT_SECONDS = 30 * 60  # 30 minutes
# only used if the Retry-After header is missing
DEFAULT_RETRY_AFTER_SECONDS: float = 1


# according to the docs, these are the only permitted completed statues
_SUCCEEDED_STATUS = "Succeeded"
_FAILED_STATUSES = ("Failed", "Canceled")

_POLL_SCHEMES = Literal[
    "AsyncOperationJsonStatus", "LocationStatusCode", "GetProvisioningState"
]


@dataclasses.dataclass(frozen=True)
class _PreparedPollRequest:
    retry_after: float
    poll_request_args: Dict[str, Any]


def _prepare_poll_request(
    poll_scheme: str,
    prev_poll_request: Optional[_PreparedPollRequest],
    response: aiohttp.ClientResponse,
    response_json: Any,
    token: str,
    original_url: str,
    api_version: str,
) -> Optional[_PreparedPollRequest]:
    """
    If prev_poll_request is None, this means that this is the initial response, which is
    treated differently in some schemes.
    """

    # common to all poll schemes

    retry_after_str = response.headers.get("Retry-After")
    if retry_after_str is None:
        retry_after = DEFAULT_RETRY_AFTER_SECONDS
    else:
        try:
            retry_after = float(retry_after_str)
        except ValueError:
            retry_after = DEFAULT_RETRY_AFTER_SECONDS

    headers = {"Authorization": f"Bearer {token}"}

    request_id = response.headers.get("x-ms-request-id")
    if request_id:
        headers["request-id"] = request_id

    if poll_scheme == "AsyncOperationJsonStatus":
        # check to see if further polling is needed
        if prev_poll_request is None:
            # on the initial response, status codes are used to indicate whether polling
            # is needed
            if response.status not in (201, 202):
                return None
        else:
            # on polling response, status code will always be 200, and we need to check
            # the returned content
            if response_json["status"] == "Succeeded":
                return None
            if response_json["status"] in _FAILED_STATUSES:
                raise AzureRestApiError(
                    response.status, response_json["status"], "Failure while polling"
                )

        # prepare the poll request if polling is needed
        if prev_poll_request is None:
            url = response.headers["Azure-AsyncOperation"]
        else:
            # the Azure-AsyncOperation header will only be provided on the initial
            # response. After that, we have to "remember" it from the previous call
            url = prev_poll_request.poll_request_args["url"]
        return _PreparedPollRequest(retry_after, {"url": url, "headers": headers})
    elif poll_scheme == "LocationStatusCode":
        if response.status not in (201, 202):
            return None

        return _PreparedPollRequest(
            retry_after, {"url": response.headers["Location"], "headers": headers}
        )
    elif poll_scheme == "GetProvisioningState":
        provisioning_state = response_json["properties"]["provisioningState"]
        if provisioning_state == _SUCCEEDED_STATUS:
            return None
        elif provisioning_state in _FAILED_STATUSES:
            raise AzureRestApiError(
                response.status, provisioning_state, "Failure while polling"
            )

        return _PreparedPollRequest(
            retry_after,
            {
                "url": original_url,
                "params": {"api-version": api_version},
                "headers": headers,
            },
        )
    else:
        raise ValueError(f"Unexpected poll_scheme {poll_scheme}")


async def azure_rest_api_poll(
    method: str,
    url_path: str,
    api_version: str,
    poll_scheme: _POLL_SCHEMES,
    *,
    query_parameters: Optional[Dict[str, str]] = None,
    json_content: Any = None,
    total_timeout_seconds: int = DEFAULT_TOTAL_POLL_TIMEOUT_SECONDS,
) -> Tuple[Any, Optional[Coroutine[Any, Any, Any]]]:
    """
    Supports any Azure REST API that requires polling. Polling in the Azure REST API is
    very complicated:

    https://docs.microsoft.com/en-us/azure/azure-resource-manager/management/async-operations#create-storage-account-202-with-location-and-retry-after
    The documentation is pretty fuzzy, but it seems like there are in practice 3 polling
    schemes in the Azure API:
    - AsyncOperationJsonStatus: Initial response is 201, has an Azure-AsyncOperation
      header that points us to the polling url, and the content contains information
      about the resource. The polling URL always returns 200, and indicates completion
      by response_json["status"] being a completed status. If polling isn't needed, the
      initial response will be 200. If the documentation for a method says that it can
      return 201, it may use this scheme:
      https://docs.microsoft.com/en-us/rest/api/compute/virtual-machines/create-or-update
      But other methods like
      https://docs.microsoft.com/en-us/rest/api/resources/resource-groups/create-or-update
      return 201, but do not include an Azure-AsyncOperation header. If anything, they
      theoretically fall under the GetProvisioningState scheme, but that example seems
      to always return a completed provisioningState on the initial call.
    - LocationStatusCode: Initial response is 202, has a Location header that points us
      to the polling url, and the content is empty. The polling URL will return 202 with
      empty content if not complete, and 200 once complete with the content containing
      information about the resource. If polling isn't needed, the initial response will
      be 200. It seems like the documentation for a method will say that it can return
      202 if this scheme is used:
      https://docs.microsoft.com/en-us/rest/api/resources/resource-groups/delete
    - GetProvisioningState: Initial response is 200 or 201, does not have any special
      headers. response_json["properties"]["provisioningState"] is populated. We can
      repeatedly call GET on the resource to check this same provisioningState value.
      (This one doesn't seem to be documented, but we've seen it happen in practice.)

    If the initial response indicates completion (i.e. no polling required), then this
    function will return (body of initial call, None). If polling is required, this will
    return (body of initial call, Awaitable[body of final polling call]). Use this with
    wait_for_poll to instead always get back (body of initial call, body of final
    polling call if polling was required).

    total_timeout_seconds indicates how long from start to finish we are willing to wait
    and poll for.
    """

    t0 = time.time()

    token = await get_token()
    prepared_request = _prepare_request(
        method,
        url_path,
        api_version,
        token,
        query_parameters=query_parameters,
        json_content=json_content,
    )

    async with aiohttp.request(**prepared_request) as response:
        await raise_for_status(response)
        response_json = await _return_response(response)

        prepared_poll_request = _prepare_poll_request(
            poll_scheme,
            None,
            response,
            response_json,
            token,
            prepared_request["url"],
            api_version,
        )
        if prepared_poll_request is None:
            # response indicates that there's no need for polling
            return response_json, None

        return response_json, _azure_rest_api_poll_continuation(
            prepared_poll_request,
            poll_scheme,
            token,
            t0,
            total_timeout_seconds,
            prepared_request["url"],
            api_version,
        )


async def _azure_rest_api_poll_continuation(
    prepared_poll_request: _PreparedPollRequest,
    poll_scheme: _POLL_SCHEMES,
    token: str,
    t0: float,
    total_timeout_seconds: int,
    original_url: str,
    api_version: str,
) -> Any:
    # https://docs.microsoft.com/en-us/azure/azure-resource-manager/management/async-operations
    while time.time() - t0 < total_timeout_seconds:
        print(
            f"Waiting for a long-running Azure operation "
            f"({prepared_poll_request.retry_after}s)"
        )
        await asyncio.sleep(prepared_poll_request.retry_after)
        async with aiohttp.request(
            "GET", **prepared_poll_request.poll_request_args
        ) as response:
            await raise_for_status(response)
            response_json = await _return_response(response)

            next_poll_request = _prepare_poll_request(
                poll_scheme,
                prepared_poll_request,
                response,
                response_json,
                token,
                original_url,
                api_version,
            )
            if next_poll_request is None:
                return response_json
            prepared_poll_request = next_poll_request

    raise TimeoutError(
        f"azure_rest_api_poll to {original_url} timed out after {total_timeout_seconds}"
        " seconds"
    )


async def wait_for_poll(
    poll_result: Tuple[Any, Optional[Coroutine[Any, Any, Any]]]
) -> Tuple[Any, Any]:
    """See azure_rest_api_poll"""
    first_result, continuation = poll_result
    if continuation is not None:
        second_result = await continuation
    else:
        second_result = None

    return first_result, second_result


# these functions really belong in azure_identity.py, but they depend on azure_rest_api
# and friends


async def get_tenant_id() -> str:
    """Gets the tenant id corresponding to the subscription from get_subscription_id"""
    # transitively can modify CACHED_AZURE_VARIABLES.subscription_id as well

    if CACHED_AZURE_VARIABLES.tenant_id is None:
        if CACHED_AZURE_VARIABLES.subscription_id is None:
            await get_subscription_id()

        # it's possible that get_subscription_id populated _TENANT_ID in which case we
        # don't need to do anything else
        if CACHED_AZURE_VARIABLES.tenant_id is None:
            # in this case we need to look up the tenant id from the subscription id
            CACHED_AZURE_VARIABLES.tenant_id = (
                await azure_rest_api(
                    "GET",
                    f"subscriptions/{CACHED_AZURE_VARIABLES.subscription_id}",
                    "2019-06-01",
                )
            )["tenantId"]
    return CACHED_AZURE_VARIABLES.tenant_id


async def get_subscription_id() -> str:
    if CACHED_AZURE_VARIABLES.subscription_id is None:
        # first, get_token might populate the subscription and tenant ids, because the
        # CLI token comes with that information
        await get_token()
        if CACHED_AZURE_VARIABLES.subscription_id is None:
            subscriptions = []
            async for page in azure_rest_api_paged(
                "GET", "subscriptions", "2021-01-01"
            ):
                for sub in page["value"]:
                    if sub["state"] == "Enabled":
                        subscriptions.append(sub)

            if len(subscriptions) > 1:
                raise ValueError(
                    "Please specify a subscription via the "
                    "AZURE_SUBSCRIPTION_ID environment variable from among the "
                    "available subscription ids: "
                    + ", ".join([sub["subscriptionId"] for sub in subscriptions])
                )
            elif len(subscriptions) == 0:
                raise ValueError("There are no subscriptions available")
            else:
                CACHED_AZURE_VARIABLES.subscription_id = cast(
                    str, subscriptions[0]["subscriptionId"]
                )
                CACHED_AZURE_VARIABLES.tenant_id = cast(
                    str, subscriptions[0]["tenantId"]
                )

    return CACHED_AZURE_VARIABLES.subscription_id


def get_subscription_id_sync() -> str:
    if CACHED_AZURE_VARIABLES.subscription_id is None:
        # first, get_token might populate the subscription and tenant ids, because the
        # CLI token comes with that information
        get_token_sync()
        if CACHED_AZURE_VARIABLES.subscription_id is None:
            subscriptions = []
            for page in azure_rest_api_paged_sync("GET", "subscriptions", "2021-01-01"):
                for sub in page["value"]:
                    if sub["state"] == "Enabled":
                        subscriptions.append(sub)

            if len(subscriptions) > 1:
                raise ValueError(
                    "Please specify a subscription via the "
                    "AZURE_SUBSCRIPTION_ID environment variable from among the "
                    "available subscription ids: "
                    + ", ".join([sub["subscriptionId"] for sub in subscriptions])
                )
            elif len(subscriptions) == 0:
                raise ValueError("There are no subscriptions available")
            else:
                CACHED_AZURE_VARIABLES.subscription_id = cast(
                    str, subscriptions[0]["subscriptionId"]
                )
                CACHED_AZURE_VARIABLES.tenant_id = cast(
                    str, subscriptions[0]["tenantId"]
                )

    return CACHED_AZURE_VARIABLES.subscription_id


# this is kind of a bit of miscellaneous functionality


def parse_azure_timestamp(s: str) -> datetime.datetime:
    """
    Azure Table timestamps are always in UTC.

    This isn't a simple parse because Azure Table timestamps have more than 1/10^6
    second precision, so we need to chop off the extra digits before python's
    datetime.strptime will work.
    """
    main, sep, frac = s.rpartition(".")
    # frac will be e.g. 0.3535722Z
    if sep != ".":
        raise ValueError(f"Unable to parse Timestamp from Azure table: {s}")

    if frac.endswith("Z"):
        frac = frac[:-1]  # drop the Z, so 0.3535722
    elif frac.endswith("+00:00"):
        frac = frac[:-6]
    else:
        raise ValueError(f"Unable to parse Timestamp from Azure table: {s}")

    frac_rounded = round(float(f"0.{frac}"), 6)  # round to 6 digits, so 0.353572
    frac = str(frac_rounded)[2:]  # drop the 0., so 353572

    # recombine and parse
    return datetime.datetime.strptime(f"{main}.{frac}", "%Y-%m-%dT%H:%M:%S.%f")



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class _EnvironmentType:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _EnvironmentTypeEnumTypeWrapper(
    google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[
        _EnvironmentType.ValueType
    ],
    builtins.type,
):  # noqa: F821
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    DEFAULT: _EnvironmentType.ValueType  # 0
    """Reserved, not used"""
    CONDA: _EnvironmentType.ValueType  # 1
    PIP: _EnvironmentType.ValueType  # 2
    POETRY: _EnvironmentType.ValueType  # 3

class EnvironmentType(_EnvironmentType, metaclass=_EnvironmentTypeEnumTypeWrapper): ...

DEFAULT: EnvironmentType.ValueType  # 0
"""Reserved, not used"""
CONDA: EnvironmentType.ValueType  # 1
PIP: EnvironmentType.ValueType  # 2
POETRY: EnvironmentType.ValueType  # 3
global___EnvironmentType = EnvironmentType

class StringPair(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    KEY_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    key: builtins.str
    value: builtins.str
    def __init__(
        self,
        *,
        key: builtins.str = ...,
        value: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self, field_name: typing_extensions.Literal["key", b"key", "value", b"value"]
    ) -> None: ...

global___StringPair = StringPair

class ServerAvailableFolder(google.protobuf.message.Message):
    """Represents a folder (or folders) that contain code that the meadowrun server can
    access directly
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CODE_PATHS_FIELD_NUMBER: builtins.int
    @property
    def code_paths(
        self,
    ) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """code_paths[0] will be set as the working directory, and all code_paths will be
        added to the PYTHONPATH. These code_paths must "make sense" on the machine where
        the meadowrun agent is running. One typical use case for this is that the
        meadowrun agents have access to a shared filesystem where code has been deployed.
        Order matters as usual for PYTHONPATH. Another typical use case is to provide no
        code_paths because all of the code needed is already specified in the
        interpreter_deployment
        """
    def __init__(
        self,
        *,
        code_paths: collections.abc.Iterable[builtins.str] | None = ...,
    ) -> None: ...
    def ClearField(
        self, field_name: typing_extensions.Literal["code_paths", b"code_paths"]
    ) -> None: ...

global___ServerAvailableFolder = ServerAvailableFolder

class CodeZipFile(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    URL_FIELD_NUMBER: builtins.int
    CODE_PATHS_FIELD_NUMBER: builtins.int
    CWD_PATH_FIELD_NUMBER: builtins.int
    url: builtins.str
    """A single zip file that contains code which will be unzipped and made available.
    All code_paths are relative to the zip file's root, and will be added to the
    PYTHONPATH. Order matters as usual for PYTHONPATH.
    Current working directory path is also relative to zip root, and should be the folder
    where the command/function is exectued on the remote side.
    The zip file is located by URL.
    For s3 this is s3://bucket-name/key-name
    For files this is file://path/to/file
    """
    @property
    def code_paths(
        self,
    ) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[
        builtins.str
    ]: ...
    cwd_path: builtins.str
    def __init__(
        self,
        *,
        url: builtins.str = ...,
        code_paths: collections.abc.Iterable[builtins.str] | None = ...,
        cwd_path: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "code_paths", b"code_paths", "cwd_path", b"cwd_path", "url", b"url"
        ],
    ) -> None: ...

global___CodeZipFile = CodeZipFile

class GitRepoCommit(google.protobuf.message.Message):
    """Represents a git repo at a specific commit"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    REPO_URL_FIELD_NUMBER: builtins.int
    COMMIT_FIELD_NUMBER: builtins.int
    PATH_TO_SOURCE_FIELD_NUMBER: builtins.int
    repo_url: builtins.str
    """specifies the url, will be provided to git clone, see
    https://git-scm.com/docs/git-clone
    """
    commit: builtins.str
    """specifies the commit to use, will be provided to git checkout [commit] see
    https://git-scm.com/book/en/v2/Git-Tools-Revision-Selection
    """
    path_to_source: builtins.str
    """specifies a relative path within the repo to treat as the "root" directory for
    the purposes of this deployment
    """
    def __init__(
        self,
        *,
        repo_url: builtins.str = ...,
        commit: builtins.str = ...,
        path_to_source: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "commit",
            b"commit",
            "path_to_source",
            b"path_to_source",
            "repo_url",
            b"repo_url",
        ],
    ) -> None: ...

global___GitRepoCommit = GitRepoCommit

class GitRepoBranch(google.protobuf.message.Message):
    """Represents a git repo on a specific branch. Note that this is NOT deterministic as
    the coordinator will resolve the branch to a specific commit. In order to reproduce
    any results, the code must be run with the specific commit that this resolves to, NOT
    the branch that was originally specified. This should only be used when GitRepoBranch
    cannot be resolved to a GitRepoCommit on the client.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    REPO_URL_FIELD_NUMBER: builtins.int
    BRANCH_FIELD_NUMBER: builtins.int
    PATH_TO_SOURCE_FIELD_NUMBER: builtins.int
    repo_url: builtins.str
    """specifies the url, will be provided to git clone, see
    https://git-scm.com/docs/git-clone
    """
    branch: builtins.str
    """specifies the branch to use"""
    path_to_source: builtins.str
    """specifies a relative path within the repo to treat as the "root" directory for
    the purposes of this deployment
    """
    def __init__(
        self,
        *,
        repo_url: builtins.str = ...,
        branch: builtins.str = ...,
        path_to_source: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "branch",
            b"branch",
            "path_to_source",
            b"path_to_source",
            "repo_url",
            b"repo_url",
        ],
    ) -> None: ...

global___GitRepoBranch = GitRepoBranch

class ServerAvailableInterpreter(google.protobuf.message.Message):
    """Represents an interpreter that the meadowrun server can access directly.
    interpreter_path can be set to meadowrun.config.MEADOWRUN_INTERPRETER to indicate
    that this job should run using the same interpreter that's being used to run
    meadowrun, which is only recommended for testing.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    INTERPRETER_PATH_FIELD_NUMBER: builtins.int
    interpreter_path: builtins.str
    def __init__(
        self,
        *,
        interpreter_path: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal["interpreter_path", b"interpreter_path"],
    ) -> None: ...

global___ServerAvailableInterpreter = ServerAvailableInterpreter

class ContainerAtDigest(google.protobuf.message.Message):
    """Represents a specific version (aka digest) of a container"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    REPOSITORY_FIELD_NUMBER: builtins.int
    DIGEST_FIELD_NUMBER: builtins.int
    repository: builtins.str
    """Together, repository and digest should be such that `docker pull
    [repository]@[digest]` works
    """
    digest: builtins.str
    def __init__(
        self,
        *,
        repository: builtins.str = ...,
        digest: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "digest", b"digest", "repository", b"repository"
        ],
    ) -> None: ...

global___ContainerAtDigest = ContainerAtDigest

class ContainerAtTag(google.protobuf.message.Message):
    """Represents a tag of a container. Note that this is NOT deterministic as the
    coordinator will resolve the tag to a specific digest. In order to reproduce any
    results, the code must be run with the specific digest that this resolves to, NOT the
    tag that was originally specified. This should only be used when ContainerAtTag
    cannot be resolved to a ContainerAtDigest on the client.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    REPOSITORY_FIELD_NUMBER: builtins.int
    TAG_FIELD_NUMBER: builtins.int
    repository: builtins.str
    """Together, repository and tag should be such that `docker pull [repository]:[tag]`
    works. The container should be configured so that `docker run [repository]:[tag]
    python [additional arguments]` behaves as expected
    """
    tag: builtins.str
    def __init__(
        self,
        *,
        repository: builtins.str = ...,
        tag: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "repository", b"repository", "tag", b"tag"
        ],
    ) -> None: ...

global___ContainerAtTag = ContainerAtTag

class EnvironmentSpecInCode(google.protobuf.message.Message):
    """Represents the path to an environment spec in the code deployment.
    An environment spec is a list of packages to install, in a format that
    the package manager can understand.
    It is not necessarily reproducible, e.g.
    if it doesn't contain all transitive dependencies or full version numbers.
    It is also not necessarily cross-platform.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class AdditionalSoftwareEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        value: builtins.str
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: builtins.str = ...,
        ) -> None: ...
        def ClearField(
            self,
            field_name: typing_extensions.Literal["key", b"key", "value", b"value"],
        ) -> None: ...

    ENVIRONMENT_TYPE_FIELD_NUMBER: builtins.int
    PATH_TO_SPEC_FIELD_NUMBER: builtins.int
    PYTHON_VERSION_FIELD_NUMBER: builtins.int
    ADDITIONAL_SOFTWARE_FIELD_NUMBER: builtins.int
    environment_type: global___EnvironmentType.ValueType
    path_to_spec: builtins.str
    python_version: builtins.str
    """python version is not used for CONDA environments"""
    @property
    def additional_software(
        self,
    ) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.str]:
        """right now just cuda is supported, maps software -> version requirement"""
    def __init__(
        self,
        *,
        environment_type: global___EnvironmentType.ValueType = ...,
        path_to_spec: builtins.str = ...,
        python_version: builtins.str = ...,
        additional_software: collections.abc.Mapping[builtins.str, builtins.str]
        | None = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "additional_software",
            b"additional_software",
            "environment_type",
            b"environment_type",
            "path_to_spec",
            b"path_to_spec",
            "python_version",
            b"python_version",
        ],
    ) -> None: ...

global___EnvironmentSpecInCode = EnvironmentSpecInCode

class EnvironmentSpec(google.protobuf.message.Message):
    """Represents an environment spec represented as a string.
    An environment spec is a list of packages to install, in a format that
    the package manager can understand.
    e.g. output of "conda env export"
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class AdditionalSoftwareEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        value: builtins.str
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: builtins.str = ...,
        ) -> None: ...
        def ClearField(
            self,
            field_name: typing_extensions.Literal["key", b"key", "value", b"value"],
        ) -> None: ...

    ENVIRONMENT_TYPE_FIELD_NUMBER: builtins.int
    SPEC_FIELD_NUMBER: builtins.int
    SPEC_LOCK_FIELD_NUMBER: builtins.int
    PYTHON_VERSION_FIELD_NUMBER: builtins.int
    ADDITIONAL_SOFTWARE_FIELD_NUMBER: builtins.int
    environment_type: global___EnvironmentType.ValueType
    spec: builtins.str
    spec_lock: builtins.str
    """only applicable for poetry"""
    python_version: builtins.str
    """python version is not used for CONDA environments"""
    @property
    def additional_software(
        self,
    ) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.str]:
        """right now just cuda is supported, maps software -> version requirement"""
    def __init__(
        self,
        *,
        environment_type: global___EnvironmentType.ValueType = ...,
        spec: builtins.str = ...,
        spec_lock: builtins.str = ...,
        python_version: builtins.str = ...,
        additional_software: collections.abc.Mapping[builtins.str, builtins.str]
        | None = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "additional_software",
            b"additional_software",
            "environment_type",
            b"environment_type",
            "python_version",
            b"python_version",
            "spec",
            b"spec",
            "spec_lock",
            b"spec_lock",
        ],
    ) -> None: ...

global___EnvironmentSpec = EnvironmentSpec

class ServerAvailableContainer(google.protobuf.message.Message):
    """Only recommended for testing. Represents a container image that already exists on the
    meadowrun server. Helpful for testing with locally built images that haven't been
    uploaded to a repository and don't have a digest
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    IMAGE_NAME_FIELD_NUMBER: builtins.int
    image_name: builtins.str
    def __init__(
        self,
        *,
        image_name: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self, field_name: typing_extensions.Literal["image_name", b"image_name"]
    ) -> None: ...

global___ServerAvailableContainer = ServerAvailableContainer

class PyCommandJob(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    COMMAND_LINE_FIELD_NUMBER: builtins.int
    PICKLED_CONTEXT_VARIABLES_FIELD_NUMBER: builtins.int
    @property
    def command_line(
        self,
    ) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[
        builtins.str
    ]: ...
    pickled_context_variables: builtins.bytes
    def __init__(
        self,
        *,
        command_line: collections.abc.Iterable[builtins.str] | None = ...,
        pickled_context_variables: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "command_line",
            b"command_line",
            "pickled_context_variables",
            b"pickled_context_variables",
        ],
    ) -> None: ...

global___PyCommandJob = PyCommandJob

class QualifiedFunctionName(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    MODULE_NAME_FIELD_NUMBER: builtins.int
    FUNCTION_NAME_FIELD_NUMBER: builtins.int
    module_name: builtins.str
    function_name: builtins.str
    def __init__(
        self,
        *,
        module_name: builtins.str = ...,
        function_name: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "function_name", b"function_name", "module_name", b"module_name"
        ],
    ) -> None: ...

global___QualifiedFunctionName = QualifiedFunctionName

class PyFunctionJob(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    QUALIFIED_FUNCTION_NAME_FIELD_NUMBER: builtins.int
    PICKLED_FUNCTION_FIELD_NUMBER: builtins.int
    PICKLED_FUNCTION_ARGUMENTS_FIELD_NUMBER: builtins.int
    @property
    def qualified_function_name(self) -> global___QualifiedFunctionName: ...
    pickled_function: builtins.bytes
    pickled_function_arguments: builtins.bytes
    def __init__(
        self,
        *,
        qualified_function_name: global___QualifiedFunctionName | None = ...,
        pickled_function: builtins.bytes = ...,
        pickled_function_arguments: builtins.bytes = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "function_spec",
            b"function_spec",
            "pickled_function",
            b"pickled_function",
            "qualified_function_name",
            b"qualified_function_name",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "function_spec",
            b"function_spec",
            "pickled_function",
            b"pickled_function",
            "pickled_function_arguments",
            b"pickled_function_arguments",
            "qualified_function_name",
            b"qualified_function_name",
        ],
    ) -> None: ...
    def WhichOneof(
        self, oneof_group: typing_extensions.Literal["function_spec", b"function_spec"]
    ) -> typing_extensions.Literal[
        "qualified_function_name", "pickled_function"
    ] | None: ...

global___PyFunctionJob = PyFunctionJob

class PyAgentJob(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    QUALIFIED_FUNCTION_NAME_FIELD_NUMBER: builtins.int
    PICKLED_FUNCTION_FIELD_NUMBER: builtins.int
    PICKLED_AGENT_FUNCTION_FIELD_NUMBER: builtins.int
    PICKLED_AGENT_FUNCTION_ARGUMENTS_FIELD_NUMBER: builtins.int
    @property
    def qualified_function_name(self) -> global___QualifiedFunctionName: ...
    pickled_function: builtins.bytes
    pickled_agent_function: builtins.bytes
    pickled_agent_function_arguments: builtins.bytes
    def __init__(
        self,
        *,
        qualified_function_name: global___QualifiedFunctionName | None = ...,
        pickled_function: builtins.bytes = ...,
        pickled_agent_function: builtins.bytes = ...,
        pickled_agent_function_arguments: builtins.bytes = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "function_spec",
            b"function_spec",
            "pickled_function",
            b"pickled_function",
            "qualified_function_name",
            b"qualified_function_name",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "function_spec",
            b"function_spec",
            "pickled_agent_function",
            b"pickled_agent_function",
            "pickled_agent_function_arguments",
            b"pickled_agent_function_arguments",
            "pickled_function",
            b"pickled_function",
            "qualified_function_name",
            b"qualified_function_name",
        ],
    ) -> None: ...
    def WhichOneof(
        self, oneof_group: typing_extensions.Literal["function_spec", b"function_spec"]
    ) -> typing_extensions.Literal[
        "qualified_function_name", "pickled_function"
    ] | None: ...

global___PyAgentJob = PyAgentJob

class GridTask(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TASK_ID_FIELD_NUMBER: builtins.int
    ATTEMPT_FIELD_NUMBER: builtins.int
    PICKLED_FUNCTION_ARGUMENTS_FIELD_NUMBER: builtins.int
    task_id: builtins.int
    attempt: builtins.int
    pickled_function_arguments: builtins.bytes
    def __init__(
        self,
        *,
        task_id: builtins.int = ...,
        attempt: builtins.int = ...,
        pickled_function_arguments: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "attempt",
            b"attempt",
            "pickled_function_arguments",
            b"pickled_function_arguments",
            "task_id",
            b"task_id",
        ],
    ) -> None: ...

global___GridTask = GridTask

class ContainerImage(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CONTAINER_IMAGE_AT_DIGEST_FIELD_NUMBER: builtins.int
    CONTAINER_IMAGE_AT_TAG_FIELD_NUMBER: builtins.int
    SERVER_AVAILABLE_CONTAINER_IMAGE_FIELD_NUMBER: builtins.int
    @property
    def container_image_at_digest(self) -> global___ContainerAtDigest: ...
    @property
    def container_image_at_tag(self) -> global___ContainerAtTag: ...
    @property
    def server_available_container_image(self) -> global___ServerAvailableContainer: ...
    def __init__(
        self,
        *,
        container_image_at_digest: global___ContainerAtDigest | None = ...,
        container_image_at_tag: global___ContainerAtTag | None = ...,
        server_available_container_image: global___ServerAvailableContainer
        | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "container_image",
            b"container_image",
            "container_image_at_digest",
            b"container_image_at_digest",
            "container_image_at_tag",
            b"container_image_at_tag",
            "server_available_container_image",
            b"server_available_container_image",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "container_image",
            b"container_image",
            "container_image_at_digest",
            b"container_image_at_digest",
            "container_image_at_tag",
            b"container_image_at_tag",
            "server_available_container_image",
            b"server_available_container_image",
        ],
    ) -> None: ...
    def WhichOneof(
        self,
        oneof_group: typing_extensions.Literal["container_image", b"container_image"],
    ) -> typing_extensions.Literal[
        "container_image_at_digest",
        "container_image_at_tag",
        "server_available_container_image",
    ] | None: ...

global___ContainerImage = ContainerImage

class Job(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    JOB_ID_FIELD_NUMBER: builtins.int
    JOB_FRIENDLY_NAME_FIELD_NUMBER: builtins.int
    SERVER_AVAILABLE_FOLDER_FIELD_NUMBER: builtins.int
    GIT_REPO_COMMIT_FIELD_NUMBER: builtins.int
    GIT_REPO_BRANCH_FIELD_NUMBER: builtins.int
    CODE_ZIP_FILE_FIELD_NUMBER: builtins.int
    SERVER_AVAILABLE_INTERPRETER_FIELD_NUMBER: builtins.int
    CONTAINER_AT_DIGEST_FIELD_NUMBER: builtins.int
    CONTAINER_AT_TAG_FIELD_NUMBER: builtins.int
    SERVER_AVAILABLE_CONTAINER_FIELD_NUMBER: builtins.int
    ENVIRONMENT_SPEC_IN_CODE_FIELD_NUMBER: builtins.int
    ENVIRONMENT_SPEC_FIELD_NUMBER: builtins.int
    SIDECAR_CONTAINERS_FIELD_NUMBER: builtins.int
    ENVIRONMENT_VARIABLES_FIELD_NUMBER: builtins.int
    RESULT_HIGHEST_PICKLE_PROTOCOL_FIELD_NUMBER: builtins.int
    PY_COMMAND_FIELD_NUMBER: builtins.int
    PY_FUNCTION_FIELD_NUMBER: builtins.int
    PY_AGENT_FIELD_NUMBER: builtins.int
    CREDENTIALS_SOURCES_FIELD_NUMBER: builtins.int
    PORTS_FIELD_NUMBER: builtins.int
    USES_GPU_FIELD_NUMBER: builtins.int
    job_id: builtins.str
    """job_id uniquely identifies this request to avoid duplicates and for getting the
    results later. Make sure job_id is unique! Multiple requests with the same job_id
    will be treated as duplicates even if all of the other parameters are different.
    Also, job_id may only use string.ascii_letters, numbers, ., -, and _.
    """
    job_friendly_name: builtins.str
    @property
    def server_available_folder(self) -> global___ServerAvailableFolder: ...
    @property
    def git_repo_commit(self) -> global___GitRepoCommit: ...
    @property
    def git_repo_branch(self) -> global___GitRepoBranch: ...
    @property
    def code_zip_file(self) -> global___CodeZipFile: ...
    @property
    def server_available_interpreter(self) -> global___ServerAvailableInterpreter: ...
    @property
    def container_at_digest(self) -> global___ContainerAtDigest:
        """The container specified should be such that running `docker run
        [repository]@[digest] python --version` works. Currently only works with
        Linux containers. If code_deployment specifies any code folders, they will be
        mounted in the container as /meadowrun/code0, /meadowrun/code1, etc.
        """
    @property
    def container_at_tag(self) -> global___ContainerAtTag: ...
    @property
    def server_available_container(self) -> global___ServerAvailableContainer: ...
    @property
    def environment_spec_in_code(self) -> global___EnvironmentSpecInCode: ...
    @property
    def environment_spec(self) -> global___EnvironmentSpec: ...
    @property
    def sidecar_containers(
        self,
    ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[
        global___ContainerImage
    ]: ...
    @property
    def environment_variables(
        self,
    ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[
        global___StringPair
    ]: ...
    result_highest_pickle_protocol: builtins.int
    """result_highest_pickle_protocol tells the remote code what the highest pickle
    protocol we can read on this end is which will help it determine what pickle
    protocol to use to send back results. This should almost always be set to
    pickle.HIGHEST_PROTOCOL in the calling python process
    """
    @property
    def py_command(self) -> global___PyCommandJob: ...
    @property
    def py_function(self) -> global___PyFunctionJob: ...
    @property
    def py_agent(self) -> global___PyAgentJob: ...
    @property
    def credentials_sources(
        self,
    ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[
        global___CredentialsSourceMessage
    ]: ...
    @property
    def ports(
        self,
    ) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[
        builtins.str
    ]: ...
    uses_gpu: builtins.bool
    def __init__(
        self,
        *,
        job_id: builtins.str = ...,
        job_friendly_name: builtins.str = ...,
        server_available_folder: global___ServerAvailableFolder | None = ...,
        git_repo_commit: global___GitRepoCommit | None = ...,
        git_repo_branch: global___GitRepoBranch | None = ...,
        code_zip_file: global___CodeZipFile | None = ...,
        server_available_interpreter: global___ServerAvailableInterpreter | None = ...,
        container_at_digest: global___ContainerAtDigest | None = ...,
        container_at_tag: global___ContainerAtTag | None = ...,
        server_available_container: global___ServerAvailableContainer | None = ...,
        environment_spec_in_code: global___EnvironmentSpecInCode | None = ...,
        environment_spec: global___EnvironmentSpec | None = ...,
        sidecar_containers: collections.abc.Iterable[global___ContainerImage]
        | None = ...,
        environment_variables: collections.abc.Iterable[global___StringPair]
        | None = ...,
        result_highest_pickle_protocol: builtins.int = ...,
        py_command: global___PyCommandJob | None = ...,
        py_function: global___PyFunctionJob | None = ...,
        py_agent: global___PyAgentJob | None = ...,
        credentials_sources: collections.abc.Iterable[global___CredentialsSourceMessage]
        | None = ...,
        ports: collections.abc.Iterable[builtins.str] | None = ...,
        uses_gpu: builtins.bool = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "code_deployment",
            b"code_deployment",
            "code_zip_file",
            b"code_zip_file",
            "container_at_digest",
            b"container_at_digest",
            "container_at_tag",
            b"container_at_tag",
            "environment_spec",
            b"environment_spec",
            "environment_spec_in_code",
            b"environment_spec_in_code",
            "git_repo_branch",
            b"git_repo_branch",
            "git_repo_commit",
            b"git_repo_commit",
            "interpreter_deployment",
            b"interpreter_deployment",
            "job_spec",
            b"job_spec",
            "py_agent",
            b"py_agent",
            "py_command",
            b"py_command",
            "py_function",
            b"py_function",
            "server_available_container",
            b"server_available_container",
            "server_available_folder",
            b"server_available_folder",
            "server_available_interpreter",
            b"server_available_interpreter",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "code_deployment",
            b"code_deployment",
            "code_zip_file",
            b"code_zip_file",
            "container_at_digest",
            b"container_at_digest",
            "container_at_tag",
            b"container_at_tag",
            "credentials_sources",
            b"credentials_sources",
            "environment_spec",
            b"environment_spec",
            "environment_spec_in_code",
            b"environment_spec_in_code",
            "environment_variables",
            b"environment_variables",
            "git_repo_branch",
            b"git_repo_branch",
            "git_repo_commit",
            b"git_repo_commit",
            "interpreter_deployment",
            b"interpreter_deployment",
            "job_friendly_name",
            b"job_friendly_name",
            "job_id",
            b"job_id",
            "job_spec",
            b"job_spec",
            "ports",
            b"ports",
            "py_agent",
            b"py_agent",
            "py_command",
            b"py_command",
            "py_function",
            b"py_function",
            "result_highest_pickle_protocol",
            b"result_highest_pickle_protocol",
            "server_available_container",
            b"server_available_container",
            "server_available_folder",
            b"server_available_folder",
            "server_available_interpreter",
            b"server_available_interpreter",
            "sidecar_containers",
            b"sidecar_containers",
            "uses_gpu",
            b"uses_gpu",
        ],
    ) -> None: ...
    @typing.overload
    def WhichOneof(
        self,
        oneof_group: typing_extensions.Literal["code_deployment", b"code_deployment"],
    ) -> typing_extensions.Literal[
        "server_available_folder", "git_repo_commit", "git_repo_branch", "code_zip_file"
    ] | None: ...
    @typing.overload
    def WhichOneof(
        self,
        oneof_group: typing_extensions.Literal[
            "interpreter_deployment", b"interpreter_deployment"
        ],
    ) -> typing_extensions.Literal[
        "server_available_interpreter",
        "container_at_digest",
        "container_at_tag",
        "server_available_container",
        "environment_spec_in_code",
        "environment_spec",
    ] | None: ...
    @typing.overload
    def WhichOneof(
        self, oneof_group: typing_extensions.Literal["job_spec", b"job_spec"]
    ) -> typing_extensions.Literal["py_command", "py_function", "py_agent"] | None: ...

global___Job = Job

class ProcessState(google.protobuf.message.Message):
    """Represents the state of a process, can apply to a job or a grid task"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _ProcessStateEnum:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _ProcessStateEnumEnumTypeWrapper(
        google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[
            ProcessState._ProcessStateEnum.ValueType
        ],
        builtins.type,
    ):  # noqa: F821
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        DEFAULT: ProcessState._ProcessStateEnum.ValueType  # 0
        """Reserved, not used"""
        RUN_REQUESTED: ProcessState._ProcessStateEnum.ValueType  # 1
        """These states represent a job that is "in progress"

        The meadowrun coordinator has received the Job
        """
        RUNNING: ProcessState._ProcessStateEnum.ValueType  # 2
        """The assigned agent has launched the job. pid and log_file_name will be
        populated.
        """
        SUCCEEDED: ProcessState._ProcessStateEnum.ValueType  # 3
        """These states represent a job that is "done". log_file_name, return_code, and
        one of pid/container_id will be populated unless otherwise noted.

        The job has completed normally. pickled_result may be populated.
        """
        RUN_REQUEST_FAILED: ProcessState._ProcessStateEnum.ValueType  # 4
        """There was an exception before launching the job process. pid/container_id,
        log_file_name, and return_code will not be populated. pickled_result will be
        populated with a tuple representing the python exception from the agent
        process (see PYTHON_EXCEPTION for the format).
        """
        PYTHON_EXCEPTION: ProcessState._ProcessStateEnum.ValueType  # 5
        """A python exception was thrown from the job process. pickled_result will be a
        pickled tuple (exception_type, exception_message, exception_traceback). We
        don't pickle the exception itself because it may not be unpicklable on this
        end (e.g. it involves types that don't exist in the current process' code
        base). Exceptions are by their nature unexpected, so we shouldn't expect that
        they can be unpickled on the client.
        """
        NON_ZERO_RETURN_CODE: ProcessState._ProcessStateEnum.ValueType  # 6
        """The process exited with a non-zero return code. This could mean that a
        non-python exception was thrown (e.g. in the interpreter itself, or in a C
        extension), or os.exit was called with a non-zero argument, or there was a
        python exception thrown in the meadowrun worker code.
        """
        RESOURCES_NOT_AVAILABLE: ProcessState._ProcessStateEnum.ValueType  # 7
        """We do not have any agents that are capable of running the job given its
        resource requirements. Either reduce the resource requirements of the job or
        launch agents that have enough resources.
        """
        ERROR_GETTING_STATE: ProcessState._ProcessStateEnum.ValueType  # 8
        """There was an error while reading the outputs of the process. This could mean
        that the child process somehow silently failed to write its outputs correctly
        or there was a python exception thrown in the meadowrun worker code.
        """
        UNEXPECTED_WORKER_EXIT: ProcessState._ProcessStateEnum.ValueType  # 10
        """(run_map only) The task worker process exited unexpectedly. This can mean the
        task called exit(), or it crashed.
        """
        UNKNOWN: ProcessState._ProcessStateEnum.ValueType  # 9
        """This state represents a job that is neither "done" nor "in progress"

        We do not know the job id
        """

    class ProcessStateEnum(
        _ProcessStateEnum, metaclass=_ProcessStateEnumEnumTypeWrapper
    ): ...
    DEFAULT: ProcessState.ProcessStateEnum.ValueType  # 0
    """Reserved, not used"""
    RUN_REQUESTED: ProcessState.ProcessStateEnum.ValueType  # 1
    """These states represent a job that is "in progress"

    The meadowrun coordinator has received the Job
    """
    RUNNING: ProcessState.ProcessStateEnum.ValueType  # 2
    """The assigned agent has launched the job. pid and log_file_name will be
    populated.
    """
    SUCCEEDED: ProcessState.ProcessStateEnum.ValueType  # 3
    """These states represent a job that is "done". log_file_name, return_code, and
    one of pid/container_id will be populated unless otherwise noted.

    The job has completed normally. pickled_result may be populated.
    """
    RUN_REQUEST_FAILED: ProcessState.ProcessStateEnum.ValueType  # 4
    """There was an exception before launching the job process. pid/container_id,
    log_file_name, and return_code will not be populated. pickled_result will be
    populated with a tuple representing the python exception from the agent
    process (see PYTHON_EXCEPTION for the format).
    """
    PYTHON_EXCEPTION: ProcessState.ProcessStateEnum.ValueType  # 5
    """A python exception was thrown from the job process. pickled_result will be a
    pickled tuple (exception_type, exception_message, exception_traceback). We
    don't pickle the exception itself because it may not be unpicklable on this
    end (e.g. it involves types that don't exist in the current process' code
    base). Exceptions are by their nature unexpected, so we shouldn't expect that
    they can be unpickled on the client.
    """
    NON_ZERO_RETURN_CODE: ProcessState.ProcessStateEnum.ValueType  # 6
    """The process exited with a non-zero return code. This could mean that a
    non-python exception was thrown (e.g. in the interpreter itself, or in a C
    extension), or os.exit was called with a non-zero argument, or there was a
    python exception thrown in the meadowrun worker code.
    """
    RESOURCES_NOT_AVAILABLE: ProcessState.ProcessStateEnum.ValueType  # 7
    """We do not have any agents that are capable of running the job given its
    resource requirements. Either reduce the resource requirements of the job or
    launch agents that have enough resources.
    """
    ERROR_GETTING_STATE: ProcessState.ProcessStateEnum.ValueType  # 8
    """There was an error while reading the outputs of the process. This could mean
    that the child process somehow silently failed to write its outputs correctly
    or there was a python exception thrown in the meadowrun worker code.
    """
    UNEXPECTED_WORKER_EXIT: ProcessState.ProcessStateEnum.ValueType  # 10
    """(run_map only) The task worker process exited unexpectedly. This can mean the
    task called exit(), or it crashed.
    """
    UNKNOWN: ProcessState.ProcessStateEnum.ValueType  # 9
    """This state represents a job that is neither "done" nor "in progress"

    We do not know the job id
    """

    STATE_FIELD_NUMBER: builtins.int
    PID_FIELD_NUMBER: builtins.int
    CONTAINER_ID_FIELD_NUMBER: builtins.int
    LOG_FILE_NAME_FIELD_NUMBER: builtins.int
    PICKLED_RESULT_FIELD_NUMBER: builtins.int
    RETURN_CODE_FIELD_NUMBER: builtins.int
    MAX_MEMORY_USED_GB_FIELD_NUMBER: builtins.int
    state: global___ProcessState.ProcessStateEnum.ValueType
    pid: builtins.int
    container_id: builtins.str
    log_file_name: builtins.str
    pickled_result: builtins.bytes
    return_code: builtins.int
    max_memory_used_gb: builtins.float
    def __init__(
        self,
        *,
        state: global___ProcessState.ProcessStateEnum.ValueType = ...,
        pid: builtins.int = ...,
        container_id: builtins.str = ...,
        log_file_name: builtins.str = ...,
        pickled_result: builtins.bytes = ...,
        return_code: builtins.int = ...,
        max_memory_used_gb: builtins.float = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "container_id",
            b"container_id",
            "log_file_name",
            b"log_file_name",
            "max_memory_used_gb",
            b"max_memory_used_gb",
            "pickled_result",
            b"pickled_result",
            "pid",
            b"pid",
            "return_code",
            b"return_code",
            "state",
            b"state",
        ],
    ) -> None: ...

global___ProcessState = ProcessState

class JobStateUpdate(google.protobuf.message.Message):
    """For updating the state of a job"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    JOB_ID_FIELD_NUMBER: builtins.int
    PROCESS_STATE_FIELD_NUMBER: builtins.int
    job_id: builtins.str
    @property
    def process_state(self) -> global___ProcessState: ...
    def __init__(
        self,
        *,
        job_id: builtins.str = ...,
        process_state: global___ProcessState | None = ...,
    ) -> None: ...
    def HasField(
        self, field_name: typing_extensions.Literal["process_state", b"process_state"]
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "job_id", b"job_id", "process_state", b"process_state"
        ],
    ) -> None: ...

global___JobStateUpdate = JobStateUpdate

class GridTaskStateResponse(google.protobuf.message.Message):
    """For getting the state of a grid task"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TASK_ID_FIELD_NUMBER: builtins.int
    ATTEMPT_FIELD_NUMBER: builtins.int
    PROCESS_STATE_FIELD_NUMBER: builtins.int
    task_id: builtins.int
    attempt: builtins.int
    @property
    def process_state(self) -> global___ProcessState: ...
    def __init__(
        self,
        *,
        task_id: builtins.int = ...,
        attempt: builtins.int = ...,
        process_state: global___ProcessState | None = ...,
    ) -> None: ...
    def HasField(
        self, field_name: typing_extensions.Literal["process_state", b"process_state"]
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "attempt",
            b"attempt",
            "process_state",
            b"process_state",
            "task_id",
            b"task_id",
        ],
    ) -> None: ...

global___GridTaskStateResponse = GridTaskStateResponse

class CredentialsSourceMessage(google.protobuf.message.Message):
    """This represents a credentials source (see credentials.py)"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SERVICE_FIELD_NUMBER: builtins.int
    SERVICE_URL_FIELD_NUMBER: builtins.int
    AWS_SECRET_FIELD_NUMBER: builtins.int
    AZURE_SECRET_FIELD_NUMBER: builtins.int
    SERVER_AVAILABLE_FILE_FIELD_NUMBER: builtins.int
    KUBERNETES_SECRET_FIELD_NUMBER: builtins.int
    service: global___Credentials.Service.ValueType
    service_url: builtins.str
    @property
    def aws_secret(self) -> global___AwsSecretProto: ...
    @property
    def azure_secret(self) -> global___AzureSecretProto: ...
    @property
    def server_available_file(self) -> global___ServerAvailableFile: ...
    @property
    def kubernetes_secret(self) -> global___KubernetesSecretProto: ...
    def __init__(
        self,
        *,
        service: global___Credentials.Service.ValueType = ...,
        service_url: builtins.str = ...,
        aws_secret: global___AwsSecretProto | None = ...,
        azure_secret: global___AzureSecretProto | None = ...,
        server_available_file: global___ServerAvailableFile | None = ...,
        kubernetes_secret: global___KubernetesSecretProto | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "aws_secret",
            b"aws_secret",
            "azure_secret",
            b"azure_secret",
            "kubernetes_secret",
            b"kubernetes_secret",
            "server_available_file",
            b"server_available_file",
            "source",
            b"source",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "aws_secret",
            b"aws_secret",
            "azure_secret",
            b"azure_secret",
            "kubernetes_secret",
            b"kubernetes_secret",
            "server_available_file",
            b"server_available_file",
            "service",
            b"service",
            "service_url",
            b"service_url",
            "source",
            b"source",
        ],
    ) -> None: ...
    def WhichOneof(
        self, oneof_group: typing_extensions.Literal["source", b"source"]
    ) -> typing_extensions.Literal[
        "aws_secret", "azure_secret", "server_available_file", "kubernetes_secret"
    ] | None: ...

global___CredentialsSourceMessage = CredentialsSourceMessage

class Credentials(google.protobuf.message.Message):
    """Represents actual credentials"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _Service:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _ServiceEnumTypeWrapper(
        google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[
            Credentials._Service.ValueType
        ],
        builtins.type,
    ):  # noqa: F821
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        DEFAULT_SERVICE: Credentials._Service.ValueType  # 0
        DOCKER: Credentials._Service.ValueType  # 1
        GIT: Credentials._Service.ValueType  # 2

    class Service(_Service, metaclass=_ServiceEnumTypeWrapper): ...
    DEFAULT_SERVICE: Credentials.Service.ValueType  # 0
    DOCKER: Credentials.Service.ValueType  # 1
    GIT: Credentials.Service.ValueType  # 2

    class _Type:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _TypeEnumTypeWrapper(
        google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[
            Credentials._Type.ValueType
        ],
        builtins.type,
    ):  # noqa: F821
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        DEFAULT_TYPE: Credentials._Type.ValueType  # 0
        USERNAME_PASSWORD: Credentials._Type.ValueType  # 1
        SSH_KEY: Credentials._Type.ValueType  # 2

    class Type(_Type, metaclass=_TypeEnumTypeWrapper): ...
    DEFAULT_TYPE: Credentials.Type.ValueType  # 0
    USERNAME_PASSWORD: Credentials.Type.ValueType  # 1
    SSH_KEY: Credentials.Type.ValueType  # 2

    CREDENTIALS_FIELD_NUMBER: builtins.int
    credentials: builtins.bytes
    def __init__(
        self,
        *,
        credentials: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(
        self, field_name: typing_extensions.Literal["credentials", b"credentials"]
    ) -> None: ...

global___Credentials = Credentials

class AwsSecretProto(google.protobuf.message.Message):
    """Represents credentials stored in AWS. Must be accessible by the coordinator.
    - For credentials_type = USERNAME_PASSWORD: Expected keys are "username" and
      "password", e.g. SecretString should be like
      '{"username":"my_username","password":"my_password"}'
    - For credentials_type = SSH_KEY: Expected key is "private_key", which should contain
      the contents of the SSH private key file
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CREDENTIALS_TYPE_FIELD_NUMBER: builtins.int
    SECRET_NAME_FIELD_NUMBER: builtins.int
    credentials_type: global___Credentials.Type.ValueType
    secret_name: builtins.str
    def __init__(
        self,
        *,
        credentials_type: global___Credentials.Type.ValueType = ...,
        secret_name: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "credentials_type", b"credentials_type", "secret_name", b"secret_name"
        ],
    ) -> None: ...

global___AwsSecretProto = AwsSecretProto

class AzureSecretProto(google.protobuf.message.Message):
    """Represents credentials stored in Azure
    - For credentials_type = USERNAME_PASSWORD, expects a json like '{"username":
      "my_username", "password": "my_pasword}'
    - For credentials_type = SSH_KEY, expects a plain-text value containing the contents
      of the SSH private key file
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CREDENTIALS_TYPE_FIELD_NUMBER: builtins.int
    VAULT_NAME_FIELD_NUMBER: builtins.int
    SECRET_NAME_FIELD_NUMBER: builtins.int
    credentials_type: global___Credentials.Type.ValueType
    vault_name: builtins.str
    secret_name: builtins.str
    def __init__(
        self,
        *,
        credentials_type: global___Credentials.Type.ValueType = ...,
        vault_name: builtins.str = ...,
        secret_name: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "credentials_type",
            b"credentials_type",
            "secret_name",
            b"secret_name",
            "vault_name",
            b"vault_name",
        ],
    ) -> None: ...

global___AzureSecretProto = AzureSecretProto

class ServerAvailableFile(google.protobuf.message.Message):
    """Represents credentials in a file. Must be a file accessible by the coordinator.
    - For credentials_type = USERNAME_PASSWORD: The file must have username on the first
      line and password on the second line. Final newline character is optional
    - For credentials_type = SSH_KEY: The file should be an SSH private key file
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CREDENTIALS_TYPE_FIELD_NUMBER: builtins.int
    PATH_FIELD_NUMBER: builtins.int
    credentials_type: global___Credentials.Type.ValueType
    path: builtins.str
    def __init__(
        self,
        *,
        credentials_type: global___Credentials.Type.ValueType = ...,
        path: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "credentials_type", b"credentials_type", "path", b"path"
        ],
    ) -> None: ...

global___ServerAvailableFile = ServerAvailableFile

class KubernetesSecretProto(google.protobuf.message.Message):
    """Represents credentials stored in Kubernetes. This class doesn't really need to be a
    protobuf because when we run on Kubernetes we don't actually need to serialize the
    Job object. But we made it a protobuf anyways because all of the other types of
    secrets are protobufs.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CREDENTIALS_TYPE_FIELD_NUMBER: builtins.int
    SECRET_NAME_FIELD_NUMBER: builtins.int
    credentials_type: global___Credentials.Type.ValueType
    secret_name: builtins.str
    def __init__(
        self,
        *,
        credentials_type: global___Credentials.Type.ValueType = ...,
        secret_name: builtins.str = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "credentials_type", b"credentials_type", "secret_name", b"secret_name"
        ],
    ) -> None: ...

global___KubernetesSecretProto = KubernetesSecretProto



"""See grid_tasks_sqs.py"""
from __future__ import annotations

import asyncio
import dataclasses
import datetime
import os
import pickle
import time
import traceback
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterable,
    Iterable,
    List,
    Optional,
    Tuple,
    TypeVar,
    cast,
)

from meadowrun.azure_integration.azure_meadowrun_core import (
    ensure_meadowrun_storage_account,
    record_last_used,
)
from meadowrun.azure_integration.mgmt_functions.azure_constants import (
    _REQUEST_QUEUE_NAME_PREFIX,
    _RESULT_QUEUE_NAME_PREFIX,
    GRID_TASK_QUEUE,
    QUEUE_NAME_TIMESTAMP_FORMAT,
)
from meadowrun.azure_integration.mgmt_functions.azure_core.azure_rest_api import (
    azure_rest_api,
)
from meadowrun.azure_integration.mgmt_functions.azure_core.azure_storage_api import (
    StorageAccount,
    queue_delete_message,
    queue_receive_messages,
    queue_send_message,
)
from meadowrun.meadowrun_pb2 import GridTask, GridTaskStateResponse, ProcessState
from meadowrun.run_job_core import TaskProcessState
from meadowrun.run_job_local import WorkerMonitor, restart_worker

if TYPE_CHECKING:
    from meadowrun.run_job_core import TaskWorkerServer, WorkerProcessState

_T = TypeVar("_T")
_U = TypeVar("_U")


@dataclasses.dataclass(frozen=True)
class Queue:
    """Everything we need to be able to access a queue"""

    queue_job_id: str
    queue_name: str
    storage_account: StorageAccount


async def create_queues_for_job(job_id: str, location: str) -> Tuple[Queue, Queue]:
    storage_account = await ensure_meadowrun_storage_account(location, "create")
    now = datetime.datetime.utcnow().strftime(QUEUE_NAME_TIMESTAMP_FORMAT)
    request_queue_task = azure_rest_api(
        "PUT",
        f"{storage_account.get_path()}/queueServices/default/queues/"
        f"{_REQUEST_QUEUE_NAME_PREFIX}-{job_id}-{now}",
        "2021-09-01",
        json_content={},
    )
    result_queue_task = azure_rest_api(
        "PUT",
        f"{storage_account.get_path()}/queueServices/default/queues/"
        f"{_RESULT_QUEUE_NAME_PREFIX}-{job_id}-{now}",
        "2021-09-01",
        json_content={},
    )

    return (
        Queue(job_id, (await request_queue_task)["name"], storage_account),
        Queue(job_id, (await result_queue_task)["name"], storage_account),
    )


_WORKER_SHUTDOWN_MESSAGE = b"worker-shutdown"


async def add_tasks(request_queue: Queue, tasks: Iterable[Any]) -> None:
    await asyncio.wait(
        [
            asyncio.create_task(
                queue_send_message(
                    request_queue.storage_account,
                    request_queue.queue_name,
                    GridTask(
                        task_id=i,
                        attempt=1,
                        pickled_function_arguments=pickle.dumps(((task,), {})),
                    ).SerializeToString(),
                )
            )
            for i, task in enumerate(tasks)
        ]
    )


async def retry_task(
    request_queue: Queue, task_id: int, attempt: int, task: Any
) -> None:
    await queue_send_message(
        request_queue.storage_account,
        request_queue.queue_name,
        GridTask(
            task_id=task_id,
            attempt=attempt,
            pickled_function_arguments=pickle.dumps(((task,), {})),
        ).SerializeToString(),
    )


async def add_worker_shutdown_message(request_queue: Queue, num_workers: int) -> None:
    # TODO I think these could be sent in a single API call
    await asyncio.wait(
        [
            asyncio.create_task(
                queue_send_message(
                    request_queue.storage_account,
                    request_queue.queue_name,
                    _WORKER_SHUTDOWN_MESSAGE,
                )
            )
            for _ in range(num_workers)
        ]
    )


async def _get_task(request_queue: Queue, result_queue: Queue) -> Optional[GridTask]:
    while True:
        messages = await queue_receive_messages(
            request_queue.storage_account,
            request_queue.queue_name,
            visibility_timeout_secs=5,
        )

        if len(messages) > 0:
            break

    if len(messages) > 1:
        raise ValueError(
            "queue_receive_messages returned more than 1 message unexpectedly"
        )

    # if we got a worker done message, return None
    if messages[0].message_content == _WORKER_SHUTDOWN_MESSAGE:
        return None

    task = GridTask()
    task.ParseFromString(messages[0].message_content)

    await queue_send_message(
        result_queue.storage_account,
        result_queue.queue_name,
        GridTaskStateResponse(
            task_id=task.task_id,
            process_state=ProcessState(
                state=ProcessState.ProcessStateEnum.RUN_REQUESTED,
                # TODO needs to include public address and worker id
            ),
        ).SerializeToString(),
    )

    await queue_delete_message(
        request_queue.storage_account,
        request_queue.queue_name,
        messages[0].message_id,
        messages[0].pop_receipt,
    )

    return task


async def _complete_task(
    result_queue: Queue,
    task: GridTask,
    process_state: ProcessState,
) -> None:
    await queue_send_message(
        result_queue.storage_account,
        result_queue.queue_name,
        GridTaskStateResponse(
            task_id=task.task_id, attempt=task.attempt, process_state=process_state
        ).SerializeToString(),
    )


async def _worker_iteration(
    request_queue: Queue,
    result_queue: Queue,
    log_file_name: str,
    pid: int,
    worker_server: TaskWorkerServer,
    worker_monitor: WorkerMonitor,
) -> bool:
    task = await _get_task(request_queue, result_queue)
    if not task:
        print("Meadowrun agent: Received shutdown message. Exiting.")
        return False

    print(
        f"Meadowrun agent: About to execute task #{task.task_id}, attempt "
        f"#{task.attempt}"
    )

    try:
        worker_monitor.start_stats()
        await worker_server.send_message(task.pickled_function_arguments)
        state, result = await worker_server.receive_message()
        stats = await worker_monitor.stop_stats()

        process_state = ProcessState(
            state=ProcessState.ProcessStateEnum.SUCCEEDED
            if state == "SUCCEEDED"
            else ProcessState.ProcessStateEnum.PYTHON_EXCEPTION,
            pid=pid,
            pickled_result=result,
            return_code=0,
            log_file_name=log_file_name,
            max_memory_used_gb=stats.max_memory_used_gb,
        )

    except Exception:
        stats = await worker_monitor.stop_stats()

        process_state = ProcessState(
            state=ProcessState.ProcessStateEnum.UNEXPECTED_WORKER_EXIT,
            pid=pid,
            return_code=(await worker_monitor.try_get_return_code()) or 0,
            log_file_name=log_file_name,
            max_memory_used_gb=stats.max_memory_used_gb,
        )

        print(
            f"Meadowrun agent: Unexpected worker exit, restarting worker. In "
            f"task #{task.task_id}, attempt #{task.attempt}, {traceback.format_exc()}"
        )

        await restart_worker(worker_server, worker_monitor)

    print(
        f"Meadowrun agent: Completed task #{task.task_id}, attempt #{task.attempt}, "
        f"state {ProcessState.ProcessStateEnum.Name(process_state.state)}, max "
        f"memory {process_state.max_memory_used_gb}GB "
    )

    await _complete_task(result_queue, task, process_state)
    return True


async def worker_function(
    request_queue: Queue,
    result_queue: Queue,
    public_address: str,
    log_file_name: str,
    worker_server: TaskWorkerServer,
    worker_monitor: WorkerMonitor,
) -> None:
    pid = os.getpid()
    log_file_name = f"{public_address}:{log_file_name}"

    while await _worker_iteration(
        request_queue, result_queue, log_file_name, pid, worker_server, worker_monitor
    ):
        pass


async def get_results_unordered(
    result_queue: Queue,
    location: str,
    stop_receiving: asyncio.Event,
    all_workers_exited: asyncio.Event,
) -> AsyncIterable[Tuple[List[TaskProcessState], List[WorkerProcessState]]]:

    # TODO currently, we get back messages saying that a task is running on a particular
    # worker. We don't really do anything with these messages, but eventually we should
    # use them to react appropriately if a worker crashes unexpectedly.

    num_tasks_running, num_tasks_completed = 0, 0
    t0 = None
    updated = True
    stop_receiving_wait_task = asyncio.create_task(stop_receiving.wait())
    all_workers_exited_task = asyncio.create_task(all_workers_exited.wait())
    while not stop_receiving.is_set():
        if updated or t0 is None or time.time() - t0 > 20:
            t0 = time.time()
            updated = False
            await record_last_used(GRID_TASK_QUEUE, result_queue.queue_job_id, location)

        receive_messages_task = asyncio.create_task(
            queue_receive_messages(
                result_queue.storage_account,
                result_queue.queue_name,
                visibility_timeout_secs=10,
                num_messages=32,  # the max number of messages to get at once
            )
        )
        await asyncio.wait(
            cast(
                List[asyncio.Task],
                [
                    receive_messages_task,
                    stop_receiving_wait_task,
                    all_workers_exited_task,
                ],
            ),
            return_when=asyncio.FIRST_COMPLETED,
        )
        if stop_receiving_wait_task.done() or all_workers_exited_task.done():
            return
        else:
            results = []

            for message in receive_messages_task.result():
                updated = True

                task_result = GridTaskStateResponse()
                task_result.ParseFromString(message.message_content)

                if (
                    task_result.process_state.state
                    == ProcessState.ProcessStateEnum.RUN_REQUESTED
                ):
                    num_tasks_running += 1
                else:
                    num_tasks_running -= 1
                    num_tasks_completed += 1
                    results.append(
                        TaskProcessState(
                            task_result.task_id,
                            task_result.attempt,
                            task_result.process_state,
                        )
                    )
                await queue_delete_message(
                    result_queue.storage_account,
                    result_queue.queue_name,
                    message.message_id,
                    message.pop_receipt,
                )

            yield results, []



"""
Copied from https://github.com/talkiq/gcloud-aio/tree/storage-7.0.1/storage/gcloud/aio/storage
"""

__version__ = "7.0.1"

from meadowrun._vendor.gcloud.aio.storage.blob import Blob
from meadowrun._vendor.gcloud.aio.storage.bucket import Bucket
from meadowrun._vendor.gcloud.aio.storage.storage import SCOPES
from meadowrun._vendor.gcloud.aio.storage.storage import Storage
from meadowrun._vendor.gcloud.aio.storage.storage import StreamResponse


__all__ = ["__version__", "Blob", "Bucket", "SCOPES", "Storage", "StreamResponse"]



import base64
import dataclasses
import datetime
import hashlib
import hmac
import xml.dom.minidom
from typing import Any, AsyncIterator, Dict, Iterable, Optional, Sequence, Tuple

import aiohttp
import multidict
from .azure_exceptions import (
    raise_for_status,
)
from .azure_rest_api import (
    _return_response,
)


def _sign_string(key: str, string_to_sign: str) -> str:
    return base64.b64encode(
        hmac.HMAC(
            base64.b64decode(key), string_to_sign.encode("utf-8"), hashlib.sha256
        ).digest()
    ).decode("utf-8")


def _get_now_rfc1123() -> str:
    """
    This is a specific datetime format required for authentication with the storage APIs
    """
    # Copied from
    # https://github.com/Azure/azure-sdk-for-python/blob/1d5096eb1bc8cbd77223ecc7a628738a5f88751c/sdk/storage/azure-storage-file-datalake/azure/storage/filedatalake/_serialize.py#L45
    dt = datetime.datetime.utcnow()

    weekday = ["Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"][dt.weekday()]
    month = [
        "Jan",
        "Feb",
        "Mar",
        "Apr",
        "May",
        "Jun",
        "Jul",
        "Aug",
        "Sep",
        "Oct",
        "Nov",
        "Dec",
    ][dt.month - 1]
    return "%s, %02d %s %04d %02d:%02d:%02d GMT" % (
        weekday,
        dt.day,
        month,
        dt.year,
        dt.hour,
        dt.minute,
        dt.second,
    )


def table_key_url(table_name: str, partition_key: str, row_key: str) -> str:
    """Creates the url path for the combination of a table and a key"""
    partition_key = partition_key.replace("'", "''")
    row_key = row_key.replace("'", "''")
    return f"{table_name}(PartitionKey='{partition_key}',RowKey='{row_key}')"


@dataclasses.dataclass
class StorageAccount:
    name: str
    key: str
    path: Optional[str]

    def get_path(self) -> str:
        """
        When we construct a StorageAccount normally, it's easy and useful to include the
        path. When we construct a StorageAccount in a management function, it's hard and
        not useful to include the path so we don't do it.
        """
        if self.path is None:
            raise ValueError("Requested StorageAccount.path, but it was not available")
        return self.path


def _get_default_table_api_headers(
    storage_account: StorageAccount, url_path: str
) -> Dict[str, str]:
    """Creates default headers, including importantly the signed Authorization header"""

    headers = {
        "x-ms-version": "2019-02-02",
        "x-ms-date": _get_now_rfc1123(),
        "DataServiceVersion": "3.0",
        "Accept": "application/json;odata=minimalmetadata",
    }
    # https://docs.microsoft.com/en-us/rest/api/storageservices/authorize-with-shared-key
    # based on
    # https://github.com/Azure/azure-sdk-for-python/blob/83964018f39b7702659d208cd2640f5eea7400fc/sdk/tables/azure-data-tables/azure/data/tables/_authentication.py
    signature = _sign_string(
        storage_account.key,
        "\n".join(
            [headers.get("x-ms-date", ""), f"/{storage_account.name}/{url_path}"]
        ),
    )
    headers["Authorization"] = f"SharedKeyLite {storage_account.name}:{signature}"
    return headers


async def azure_table_api(
    method: str,
    storage_account: StorageAccount,
    url_path: str,
    *,
    query_parameters: Optional[Dict[str, str]] = None,
    json_content: Any = None,
    additional_headers: Optional[Dict[str, str]] = None,
) -> Any:
    """Supports any Azure Table API"""
    headers = _get_default_table_api_headers(storage_account, url_path)
    if additional_headers:
        headers.update(additional_headers)
    async with aiohttp.request(
        method,
        f"https://{storage_account.name}.table.core.windows.net/{url_path}",
        params=query_parameters,
        headers=headers,
        json=json_content,
    ) as response:
        await raise_for_status(response)
        return await _return_response(response)


_CONTINUATION_HEADER_PREFIX = "x-ms-continuation-"
_CONTINUATION_HEADER_PREFIX_LENGTH = len(_CONTINUATION_HEADER_PREFIX)


def _get_page_parameters(headers: multidict.CIMultiDictProxy[str]) -> Dict[str, str]:
    return {
        key[_CONTINUATION_HEADER_PREFIX_LENGTH:]: value
        for key, value in headers.items()
        if key.startswith(_CONTINUATION_HEADER_PREFIX)
    }


async def azure_table_api_paged(
    method: str,
    storage_account: StorageAccount,
    url_path: str,
    *,
    query_parameters: Optional[Dict[str, str]] = None,
) -> AsyncIterator[Any]:
    """
    Supports Azure Table APIs that require paging, as per
    https://docs.microsoft.com/en-us/rest/api/storageservices/query-timeout-and-pagination
    """

    parameters = {"$top": "1000"}
    if query_parameters:
        parameters.update(query_parameters)

    async with aiohttp.request(
        method,
        f"https://{storage_account.name}.table.core.windows.net/{url_path}",
        params=parameters,
        headers=_get_default_table_api_headers(storage_account, url_path),
    ) as response:
        await raise_for_status(response)
        yield await _return_response(response)

        page_parameters = _get_page_parameters(response.headers)

    while page_parameters:
        # a bit sloppy to modify in place, but turning page_parameters into
        # query_parameters here
        page_parameters.update(parameters)
        async with aiohttp.request(
            method,
            f"https://{storage_account.name}.table.core.windows.net/{url_path}",
            params=page_parameters,
            headers=_get_default_table_api_headers(storage_account, url_path),
        ) as response:
            await raise_for_status(response)
            yield await _return_response(response)

            page_parameters = _get_page_parameters(response.headers)


def _replace_linear_whitespace(s: str) -> str:
    return s.replace("\n", " ").replace("\r", " ").replace("\t", " ")


def _canonicalized_headers(headers: Dict[str, str]) -> str:
    "Produce the CanonicalizedHeaders for authorization."
    # https://docs.microsoft.com/en-us/rest/api/storageservices/authorize-with-shared-key#constructing-the-canonicalized-headers-string
    headers_to_sign = []
    for key, value in headers.items():
        key = key.lower()
        if key.startswith("x-ms-"):
            headers_to_sign.append(f"{key}:{_replace_linear_whitespace(value)}")
    return "\n".join(sorted(headers_to_sign))


async def azure_queue_api(
    method: str,
    storage_account: StorageAccount,
    url_path: str,
    *,
    query_parameters: Optional[Dict[str, str]] = None,
    xml_content: Optional[str] = None,
) -> Any:
    """Supports any Azure Queue API"""

    # headers and signing
    headers = {
        "x-ms-version": "2021-02-12",
        "x-ms-date": _get_now_rfc1123(),
        "Content-Type": "application/xml",
    }

    signature = _sign_string(
        storage_account.key,
        "\n".join(
            [
                method,
                "",  # Content-MD5, but we don't include it because it's optional
                headers["Content-Type"],
                # according to the docs, x-ms-date should be here, but the python Azure
                # SDK leaves this blank, and this seems to work
                "",
                _canonicalized_headers(headers),
                f"/{storage_account.name}/{url_path}",
            ]
        ),
    )

    headers["Authorization"] = f"SharedKeyLite {storage_account.name}:{signature}"

    # now actually make the request

    async with aiohttp.request(
        method,
        f"https://{storage_account.name}.queue.core.windows.net/{url_path}",
        params=query_parameters,
        headers=headers,
        data=xml_content,
    ) as response:
        await raise_for_status(response)
        return await response.text()


# high-level queue APIs

_QUEUE_MESSAGE_TEMPLATE = "<QueueMessage><MessageText>{}</MessageText></QueueMessage>"


async def queue_send_message(
    storage_account: StorageAccount, queue_name: str, message: bytes
) -> None:
    """Sends a message on the specified queues"""
    # https://docs.microsoft.com/en-us/rest/api/storageservices/put-message
    await azure_queue_api(
        "POST",
        storage_account,
        f"{queue_name}/messages",
        xml_content=_QUEUE_MESSAGE_TEMPLATE.format(
            base64.b64encode(message).decode("utf-8")
        ),
    )


@dataclasses.dataclass
class QueueMessage:
    message_id: str
    pop_receipt: str
    message_content: bytes


async def queue_receive_messages(
    storage_account: StorageAccount,
    queue_name: str,
    *,
    visibility_timeout_secs: Optional[int] = None,
    num_messages: Optional[int] = None,
) -> Sequence[QueueMessage]:
    """Receives 1 (the default) or more messages"""
    # https://docs.microsoft.com/en-us/rest/api/storageservices/get-messages
    query_parameters = {}
    if visibility_timeout_secs is not None:
        query_parameters["visibilitytimeout"] = str(visibility_timeout_secs)
    if num_messages is not None:
        query_parameters["numofmessages"] = str(num_messages)

    message_xml = await azure_queue_api(
        "GET",
        storage_account,
        f"{queue_name}/messages",
        query_parameters=query_parameters,
    )

    try:
        results = []
        message = xml.dom.minidom.parseString(message_xml)
        root_node = message.documentElement
        if root_node.nodeName != "QueueMessagesList":
            raise ValueError("Expected root element to be QueueMessages List")
        for message_node in root_node.childNodes:
            if message_node.nodeName == "QueueMessage":
                message_id = None
                pop_receipt = None
                message_text = None
                for attribute_node in message_node.childNodes:
                    if attribute_node.nodeName == "MessageId":
                        message_id = attribute_node.childNodes[0].nodeValue
                    elif attribute_node.nodeName == "PopReceipt":
                        pop_receipt = attribute_node.childNodes[0].nodeValue
                    elif attribute_node.nodeName == "MessageText":
                        message_text = attribute_node.childNodes[0].nodeValue

                if message_id is None:
                    raise ValueError("A QueueMessage was missing a MessageId")
                if pop_receipt is None:
                    raise ValueError("A QueueMessage was missing a PopReceipt")
                if message_text is None:
                    raise ValueError("A QueueMessage was missing a MessageText")
                results.append(
                    QueueMessage(
                        message_id,
                        pop_receipt,
                        base64.b64decode(message_text),
                    )
                )
    except Exception as e:
        raise ValueError(f"Cannot parse XML: {message_xml}") from e

    return results


async def queue_delete_message(
    storage_account: StorageAccount, queue_name: str, message_id: str, pop_receipt: str
) -> None:
    """Deletes the specified message"""
    # https://docs.microsoft.com/en-us/rest/api/storageservices/delete-message2
    await azure_queue_api(
        "DELETE",
        storage_account,
        f"{queue_name}/messages/{message_id}",
        query_parameters={"popreceipt": pop_receipt},
    )


def _get_default_blob_api_headers(
    method: str,
    storage_account: StorageAccount,
    url_path: str,
    content_type: str,
    query_parameters: Optional[Dict[str, str]],
    additional_headers: Optional[Dict[str, str]],
) -> Dict[str, str]:
    """Creates default headers, including importantly the signed Authorization header"""
    VERSION = "2021-08-06"
    request_date_time = _get_now_rfc1123()
    headers = {
        "x-ms-version": VERSION,
        "x-ms-date": request_date_time,
        "Content-Type": content_type,
    }
    if additional_headers:
        headers.update(additional_headers)

    canonicalized_resource = f"/{storage_account.name}/{url_path}"
    if query_parameters and "comp" in query_parameters:
        # the comp parameter is special and needs to be added to the canonicalized
        # resource
        canonicalized_resource += f"?comp={query_parameters['comp']}"

    # https://docs.microsoft.com/en-us/rest/api/storageservices/authorize-with-shared-key#blob-queue-and-file-services-shared-key-lite-authorization
    s_str = "\n".join(
        [
            method,
            "",  # Content-MD5
            content_type,
            "",  # Date
            # CanonicalizedHeaders
            _canonicalized_headers(headers),
            # CanonicalizedResource
            canonicalized_resource,
        ]
    )
    signature = _sign_string(storage_account.key, s_str)
    headers["Authorization"] = f"SharedKeyLite {storage_account.name}:{signature}"
    return headers


async def azure_blob_api(
    method: str,
    storage_account: StorageAccount,
    url_path: str,
    *,
    query_parameters: Optional[Dict[str, str]] = None,
    json_content: Any = None,
    binary_content: Any = None,
    additional_headers: Optional[Dict[str, str]] = None,
    ignored_status_codes: Iterable[Tuple[int, str]] = tuple(),
) -> Any:
    """
    Supports any Azure Blob Storage API. Will parse text, json, etc. based on the
    returned content type so can return any type
    """
    content_type = (
        "application/octet-stream" if binary_content is not None else "application/json"
    )
    headers = _get_default_blob_api_headers(
        method,
        storage_account,
        url_path,
        content_type,
        query_parameters,
        additional_headers,
    )

    async with aiohttp.request(
        method,
        f"https://{storage_account.name}.blob.core.windows.net/{url_path}",
        params=query_parameters,
        headers=headers,
        json=json_content,
        data=binary_content,
    ) as response:
        await raise_for_status(response, ignored_status_codes=ignored_status_codes)
        return await _return_response(response)


async def azure_blob_api_binary(
    method: str,
    storage_account: StorageAccount,
    url_path: str,
    *,
    query_parameters: Optional[Dict[str, str]] = None,
    json_content: Any = None,
    binary_content: Any = None,
    additional_headers: Optional[Dict[str, str]] = None,
    ignored_status_codes: Iterable[Tuple[int, str]] = tuple(),
) -> bytes:
    """Supports any Azure Blob Storage API. Always returns bytes"""
    content_type = (
        "application/octet-stream" if binary_content is not None else "application/json"
    )
    headers = _get_default_blob_api_headers(
        method,
        storage_account,
        url_path,
        content_type,
        query_parameters,
        additional_headers,
    )

    async with aiohttp.request(
        method,
        f"https://{storage_account.name}.blob.core.windows.net/{url_path}",
        params=query_parameters,
        headers=headers,
        json=json_content,
        data=binary_content,
    ) as response:
        await raise_for_status(response, ignored_status_codes=ignored_status_codes)
        return await response.content.read()



# taken from https://stackoverflow.com/a/39126754
from cryptography.hazmat.primitives import serialization as crypto_serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.backends import default_backend as crypto_default_backend


def generate_rsa_private_key() -> str:
    """Returns a newly generated private key"""
    return (
        rsa.generate_private_key(
            backend=crypto_default_backend(), public_exponent=65537, key_size=2048
        )
        .private_bytes(
            crypto_serialization.Encoding.PEM,
            crypto_serialization.PrivateFormat.TraditionalOpenSSL,
            crypto_serialization.NoEncryption(),
        )
        .decode("utf-8")
    )



from __future__ import annotations

import asyncio
from typing import (
    TYPE_CHECKING,
)

import asyncssh

if TYPE_CHECKING:
    from asyncssh import SSHClientConnection, SSHKey


@asyncssh.misc.async_context_manager
async def connect(host: str, username: str, private_key: SSHKey) -> SSHClientConnection:
    return await asyncssh.connect(
        host=host,
        username=username,
        known_hosts=None,
        config=None,
        client_keys=[private_key],
        # keepalive necessary otherwise client silently loses connection on long-running
        # jobs. The symptom is that e.g. run_and_print just hangs indefinitely.
        keepalive_interval=60,
    )


async def run_and_print(
    connection: asyncssh.SSHClientConnection, command: str, check: bool = True
) -> asyncssh.SSHCompletedProcess:
    """
    Runs the command, printing stdout and stderr from the remote process locally.
    result.stdout and result.stderr will be empty.
    Interrupts the remote process when the task is cancelled, by sending SIGINT signal.
    """
    # async with connection.create_process(command, stderr=asyncssh.STDOUT) as process:
    # doesn't work as well because that doesn't actively terminate the remote process.
    # In testing, calling terminate explicitly makes the remote process exit more
    # quickly.
    try:
        process = await connection.create_process(command, stderr=asyncssh.STDOUT)
        # TODO should be able to interleave stdout and stderr but this seems somewhat
        # non-trivial:
        # https://stackoverflow.com/questions/55299564/join-multiple-async-generators-in-python
        async for line in process.stdout:
            print(line, end="")

        return await process.wait(check)
    except (KeyboardInterrupt, asyncio.CancelledError):
        process.send_signal("INT")
        async for line in process.stdout:
            print(line, end="")
        # escalation: sends TERM and KILL respectively
        # process.terminate()
        # process.kill()
        # however those won't allow the remote process to react - asyncio.run in
        # particular makes sure on SIGINT to cancel and finish all tasks.
        raise
    finally:
        # what process.__aexit__ does:
        process.close()
        await process._chan.wait_closed()


async def run_and_capture(
    connection: asyncssh.SSHClientConnection, command: str, check: bool = True
) -> asyncssh.SSHCompletedProcess:
    """
    Runs the command, does not print any stdout/stderr. The output is also available as
    result.stdout and result.stderr.
    """
    return await connection.run(command, check=check)


async def write_to_file(
    connection: asyncssh.SSHClientConnection, bys: bytes, remote_path: str
) -> None:
    async with connection.start_sftp_client() as sftp:
        async with sftp.open(remote_path, "wb") as remote_file:
            await remote_file.write(bys)


async def write_text_to_file(
    connection: asyncssh.SSHClientConnection, text: str, remote_path: str
) -> None:
    await write_to_file(connection, text.encode("utf-8"), remote_path)


async def upload_file(
    connection: asyncssh.SSHClientConnection, local_path: str, remote_path: str
) -> None:
    await asyncssh.scp(
        local_path,
        (connection, remote_path),
    )


async def read_from_file(
    connection: asyncssh.SSHClientConnection, remote_path: str
) -> bytes:
    async with connection.start_sftp_client() as sftp:
        async with sftp.open(remote_path, "rb") as remote_file:
            return await remote_file.read()


async def read_text_from_file(
    connection: asyncssh.SSHClientConnection, remote_path: str
) -> str:
    return (await read_from_file(connection, remote_path)).decode("utf-8")



"""Code needed both by clean_up and vm_adjust"""

import os

from .azure_core.azure_storage_api import StorageAccount
from .azure_constants import (
    MEADOWRUN_RESOURCE_GROUP_NAME,
    MEADOWRUN_STORAGE_ACCOUNT_KEY_VARIABLE,
    MEADOWRUN_STORAGE_ACCOUNT_VARIABLE,
    MEADOWRUN_SUBSCRIPTION_ID,
)


def get_storage_account() -> StorageAccount:
    return StorageAccount(
        os.environ[MEADOWRUN_STORAGE_ACCOUNT_VARIABLE],
        os.environ[MEADOWRUN_STORAGE_ACCOUNT_KEY_VARIABLE],
        None,  # this isn't needed by mgmt_functions
    )


def get_resource_group_path() -> str:
    return (
        f"subscriptions/{os.environ[MEADOWRUN_SUBSCRIPTION_ID]}/resourceGroups"
        f"/{MEADOWRUN_RESOURCE_GROUP_NAME}"
    )



from __future__ import annotations

import re
from typing import TYPE_CHECKING, Dict, Optional, Tuple

import boto3

if TYPE_CHECKING:
    from meadowrun.instance_selection import OnDemandOrSpotType


# data manually copied from
# https://us-east-2.console.aws.amazon.com/servicequotas/home/services/ec2/quotas
_QUOTA_TO_INSTANCE_FAMILY: Dict[OnDemandOrSpotType, Dict[str, Tuple[str, ...]]] = {
    "spot": {
        "L-34B43A08": tuple("acdhimrtz"),
        "L-85EED4F7": ("dl",),
        "L-88CF9481": ("f",),
        "L-3819A6DF": ("g", "vt"),
        "L-B5D1601B": ("inf",),
        "L-7212CCBC": ("p",),
        "L-E3A00192": ("x",),
    },
    "on_demand": {
        "L-1216C47A": tuple("acdhimrtz"),
        "L-6E869C2A": ("dl",),
        "L-74FC7D96": ("f",),
        "L-DB2E81BA": ("g", "vt"),
        "L-1945791B": ("inf",),
        "L-417A185B": ("p",),
        "L-7295265B": ("x",),
        # TODO L-43DA4232 is for "high-memory" instances
    },
}

_INSTANCE_FAMILY_TO_QUOTA: Dict[OnDemandOrSpotType, Dict[str, str]] = {
    on_demand_or_spot: {
        family: quota
        for quota, families in quota_to_instance_family.items()
        for family in families
    }
    for on_demand_or_spot, quota_to_instance_family in _QUOTA_TO_INSTANCE_FAMILY.items()
}


class UnknownQuotaException(Exception):
    pass


def instance_family_from_type(instance_type: str) -> Optional[str]:
    instance_type = instance_type.lower()
    match = re.match(r"[a-z]+", instance_type)
    if match is None:
        return None
    return match.group()


_NOTE_THAT_TERMINATED_INSTANCES = (
    "(Note that terminated instances sometimes count against this limit: "
    "https://stackoverflow.com/a/54538652/908704 Also, quota increases are not granted "
    "immediately.)"
)


def get_quota_for_instance_type(
    instance_type: str, on_demand_or_spot: OnDemandOrSpotType, region_name: str
) -> Tuple[str, str, Tuple[str, ...]]:
    # this should be pretty rare, but we need to just throw an exception if we can't
    # figure out what kind of quota we're hitting
    unable_to_launch = (
        f"Unable to launch new {instance_type} {on_demand_or_spot} instances due to"
    )
    unknown_quota_exception = UnknownQuotaException(
        f"{unable_to_launch} a quota in your account. Please use the Service Quota tool"
        " in the AWS Console (https://.console.aws.amazon.com/servicequotas/home/"
        "services/ec2/quotas make sure the region is correct!) to find the appropriate"
        f" quota and increase it {_NOTE_THAT_TERMINATED_INSTANCES}."
    )

    instance_family = instance_family_from_type(instance_type)
    if instance_family is None:
        raise unknown_quota_exception
    quota_code = _INSTANCE_FAMILY_TO_QUOTA[on_demand_or_spot].get(instance_family)
    if quota_code is None:
        raise unknown_quota_exception

    quota_instance_families = _QUOTA_TO_INSTANCE_FAMILY[on_demand_or_spot][quota_code]

    current_quota_instance_families = ", ".join(quota_instance_families)
    quota_is_met_run_command_line = (
        f"This quota is currently met. Run `aws service-quotas "
        f"request-service-quota-increase --service-code ec2 --quota-code {quota_code} "
        "--desired-value X` to set the quota to X, where X is larger than the current "
        "quota."
    )

    try:
        current_quota_value = boto3.client(
            "service-quotas", region_name=region_name
        ).get_service_quota(ServiceCode="ec2", QuotaCode=quota_code)["Quota"]["Value"]
    except Exception:
        message = (
            f"{unable_to_launch} the {quota_code} quota, which limits how many CPUs you"
            f" can have across all of your {on_demand_or_spot} instances from the "
            f"{current_quota_instance_families} instance families. "
            f"{quota_is_met_run_command_line} {_NOTE_THAT_TERMINATED_INSTANCES}"
        )
    else:
        message = (
            f"{unable_to_launch} the {quota_code} quota which is set to "
            f"{current_quota_value}. This means you cannot have more than "
            f"{current_quota_value} CPUs across all of your {on_demand_or_spot} "
            f"instances from the {current_quota_instance_families} instance families. "
            f"{quota_is_met_run_command_line} {_NOTE_THAT_TERMINATED_INSTANCES}"
        )

    return message, quota_code, quota_instance_families



from __future__ import annotations

import enum


class EnvironmentSpecPrerequisites(enum.Flag):
    NONE = 0
    GIT = enum.auto()
    GOOGLE_AUTH = enum.auto()
    # AZURE_AUTH = enum.auto()  # not implemented yet


GOOGLE_AUTH_PACKAGE = "keyrings.google-artifactregistry-auth"



"""
This code will run in the generic AWS Lambda environment, so it should not import any
code outside this folder.
"""
import datetime
import os
from typing import Dict, Tuple, Any, Iterable

import boto3

from meadowrun.aws_integration.management_lambdas.ec2_alloc_stub import (
    _EC2_ALLOC_TABLE_NAME,
    _EC2_ALLOC_TAG,
    _EC2_ALLOC_TAG_VALUE,
    _INSTANCE_ID,
    _LAST_UPDATE_TIME,
    _RUNNING_JOBS,
    ignore_boto3_error_code,
)
from meadowrun.aws_integration.management_lambdas.config import (
    TERMINATE_INSTANCES_IF_IDLE_FOR_SECS,
)


# Terminate instances if they haven't run any jobs recently
_TERMINATE_INSTANCES_IF_IDLE_FOR = datetime.timedelta(
    seconds=TERMINATE_INSTANCES_IF_IDLE_FOR_SECS
)
# If we see instances running that aren't registered, we assume there is something wrong
# and they need to be terminated. However, it's possible that we happen to query between
# when an instance is launched and when it's registered. So for the first 30 seconds
# after an instance is launched, we don't terminate it even if it's not registered.
_LAUNCH_REGISTER_DELAY = datetime.timedelta(minutes=5)


def _get_ec2_alloc_table(region_name: str) -> Any:
    """
    Very similar to ensure_ec2_alloc_table, but assumes the table already exists and
    assumes we're running in an AWS lambda
    """

    # the AWS_REGION environment variable should be populated for lambdas
    db = boto3.resource("dynamodb", region_name=region_name)
    return db.Table(_EC2_ALLOC_TABLE_NAME)


def _get_registered_ec2_instances(
    region_name: str,
) -> Dict[str, Tuple[datetime.datetime, int]]:
    """
    Similar to _get_ec2_instances in ec2_alloc, but instead of getting the available
    resources, returns {instance_id): (last time a job was allocated or deallocated on
    this instance, number of currently running jobs)}
    """

    response = _get_ec2_alloc_table(region_name).scan(
        Select="SPECIFIC_ATTRIBUTES",
        ProjectionExpression=",".join([_INSTANCE_ID, _LAST_UPDATE_TIME, _RUNNING_JOBS]),
    )

    # see comment on ec2_alloc._get_ec2_instances
    if response.get("LastEvaluatedKey"):
        raise NotImplementedError(
            "We don't currently support a very large number of EC2 instances"
        )

    return {
        item[_INSTANCE_ID]: (
            datetime.datetime.fromisoformat(item[_LAST_UPDATE_TIME]).replace(
                tzinfo=datetime.timezone.utc
            ),
            len(item[_RUNNING_JOBS]),
        )
        for item in response["Items"]
    }


def _deregister_ec2_instance(
    instance_id: str, require_no_running_jobs: bool, region_name: str
) -> bool:
    """
    Deregisters an EC2 instance. If require_no_running_jobs is true, then only
    deregisters if there are no currently running jobs on the instance. If
    require_no_running_jobs is false, we will deregister the instance even if there are
    jobs running on it.
    """

    optional_args: Dict[str, Any] = {}
    if require_no_running_jobs:
        optional_args["ConditionExpression"] = f"size({_RUNNING_JOBS}) = :zero"
        optional_args["ExpressionAttributeValues"] = {":zero": 0}

    success, result = ignore_boto3_error_code(
        lambda: _get_ec2_alloc_table(region_name).delete_item(
            Key={_INSTANCE_ID: instance_id}, **optional_args
        ),
        "ConditionalCheckFailedException",
    )
    return success


_NON_TERMINATED_EC2_STATES = [
    "pending",
    "running",
    "shutting-down",
    "stopping",
    "stopped",
]


def adjust(region_name: str) -> None:
    _deregister_and_terminate_instances(
        region_name, _TERMINATE_INSTANCES_IF_IDLE_FOR, _LAUNCH_REGISTER_DELAY
    )
    # TODO this should also launch instances based on pre-provisioning policy


def _get_non_terminated_instances(ec2_resource: Any) -> Iterable[Any]:
    """Gets all meadowrun-launched EC2 instances that aren't terminated"""
    return ec2_resource.instances.filter(
        Filters=[
            {"Name": f"tag:{_EC2_ALLOC_TAG}", "Values": [_EC2_ALLOC_TAG_VALUE]},
            {"Name": "instance-state-name", "Values": _NON_TERMINATED_EC2_STATES},
        ]
    )


def _get_running_instances(ec2_resource: Any) -> Iterable[Any]:
    return ec2_resource.instances.filter(
        Filters=[
            {"Name": f"tag:{_EC2_ALLOC_TAG}", "Values": [_EC2_ALLOC_TAG_VALUE]},
            {"Name": "instance-state-name", "Values": ["running"]},
        ]
    )


def _deregister_and_terminate_instances(
    region_name: str,
    terminate_instances_if_idle_for: datetime.timedelta,
    launch_register_delay: datetime.timedelta = _LAUNCH_REGISTER_DELAY,
) -> None:
    """
    1. Compares running vs registered instances and terminates/deregisters instances
    to get running/registered instances back in sync
    2. Terminates and deregisters idle instances
    """

    ec2_resource = boto3.resource("ec2", region_name=region_name)
    non_terminated_instances = {
        instance.instance_id: instance
        for instance in _get_non_terminated_instances(ec2_resource)
    }
    running_instances = {
        instance.instance_id: instance
        for instance in _get_running_instances(ec2_resource)
    }

    registered_instances = _get_registered_ec2_instances(region_name)

    now = datetime.datetime.now(datetime.timezone.utc)

    for instance_id, (last_updated, num_jobs) in registered_instances.items():
        if instance_id not in running_instances:
            print(
                f"{instance_id} is registered but does not seem to be running, so we "
                "will deregister it"
            )
            _deregister_ec2_instance(instance_id, False, region_name)
            if instance_id in non_terminated_instances:
                # just in case
                non_terminated_instances[instance_id].terminate()
        elif num_jobs == 0 and (now - last_updated) > terminate_instances_if_idle_for:
            success = _deregister_ec2_instance(instance_id, True, region_name)
            if success:
                print(
                    f"{instance_id} is not running any jobs and has not run anything "
                    f"since {last_updated} so we will deregister and terminate it"
                )
                non_terminated_instances[instance_id].terminate()
        else:
            print(f"Letting {instance_id} continue to run--this instance is active")

    for instance_id, instance in non_terminated_instances.items():
        if instance_id not in registered_instances:
            if (now - instance.launch_time) > launch_register_delay:
                print(
                    f"{instance_id} is running but is not registered, will terminate. "
                    f"Was launched at {instance.launch_time}."
                )
                instance.terminate()
            else:
                print(
                    f"{instance_id} is running but is not registered, not terminating "
                    f"because it was launched recently at {instance.launch_time}"
                )


def lambda_handler(event: Any, context: Any) -> Dict[str, Any]:
    """The handler for AWS lambda"""
    adjust(os.environ["AWS_REGION"])
    return {"statusCode": 200, "body": ""}
